# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2015 Dan Falck <ddfalck@gmail.com>                      *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

"""Post Process command that will make use of the Output File and Post
Processor entries in PathJob"""


import FreeCAD
import FreeCADGui
import Path
from PathScripts import PathUtils
from Path.Post.Utils import FilenameGenerator
import os
from Path.Post.Processor import PostProcessor, PostProcessorFactory
from PySide import QtCore, QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

LOG_MODULE = Path.Log.thisModule()

DEBUG = False
if DEBUG:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


def _resolve_post_processor_name(job):
    Path.Log.debug("_resolve_post_processor_name()")
    if job.PostProcessor:
        valid_name = job.PostProcessor
    elif Path.Preferences.defaultPostProcessor():
        valid_name = Path.Preferences.defaultPostProcessor()
    elif FreeCAD.GuiUp:
        valid_name = DlgSelectPostProcessor().exec_()  # Ensure DlgSelectPostProcessor is defined
    else:
        valid_name = None

    if valid_name and PostProcessor.exists(valid_name):
        return valid_name
    else:
        raise ValueError(f"Post processor not identified.")


class DlgSelectPostProcessor:
    """Provide user with list of available and active post processor
    choices."""

    def __init__(self):
        self.dialog = FreeCADGui.PySideUic.loadUi(":/panels/DlgSelectPostProcessor.ui")
        firstItem = None
        for post in Path.Preferences.allEnabledPostProcessors():
            item = QtGui.QListWidgetItem(post)
            item.setFlags(QtCore.Qt.ItemFlag.ItemIsSelectable | QtCore.Qt.ItemFlag.ItemIsEnabled)
            self.dialog.lwPostProcessor.addItem(item)
            if not firstItem:
                firstItem = item
        if firstItem:
            self.dialog.lwPostProcessor.setCurrentItem(firstItem)
        else:
            self.dialog.buttonBox.button(QtGui.QDialogButtonBox.Ok).setEnabled(False)
        self.tooltips = {}
        self.dialog.lwPostProcessor.itemDoubleClicked.connect(self.dialog.accept)
        self.dialog.lwPostProcessor.setMouseTracking(True)
        self.dialog.lwPostProcessor.itemEntered.connect(self.updateTooltip)

    def updateTooltip(self, item):
        if item.text() in self.tooltips:
            tooltip = self.tooltips[item.text()]
        else:
            processor = PostProcessor.load(item.text())
            self.tooltips[item.text()] = processor.tooltip
            tooltip = processor.tooltip
        self.dialog.lwPostProcessor.setToolTip(tooltip)

    def exec_(self):
        if self.dialog.exec_() == 1:
            posts = self.dialog.lwPostProcessor.selectedItems()
            return posts[0].text()
        return None


class CommandPathPost:
    def GetResources(self):
        return {
            "Pixmap": "CAM_Post",
            "MenuText": QT_TRANSLATE_NOOP("CAM_Post", "Post Process"),
            "Accel": "P, P",
            "ToolTip": QT_TRANSLATE_NOOP("CAM_Post", "Post Processes the selected job"),
        }

    def IsActive(self):
        selected = FreeCADGui.Selection.getSelectionEx()
        if len(selected) != 1:
            return False

        selected_object = selected[0].Object
        self.candidate = PathUtils.findParentJob(selected_object)

        return self.candidate is not None

    def _write_file(self, filename, gcode, policy):
        #
        # Up to this point the postprocessors have been using "\n" as the end-of-line
        # characters in the gcode and using the process of writing out the file as a way
        # to convert the "\n" into whatever end-of-line characters match the system
        # running the postprocessor.  This can be a problem if the controller which will
        # run the gcode doesn't like the same end-of-line characters as the system that
        # ran the postprocessor to generate the gcode.
        # The refactored code base now allows for four possible types of end-of-line
        # characters in the gcode.
        #
        if len(gcode) > 1 and gcode[0:2] == "\n\n":
            # The gcode shouldn't normally start with "\n\n".
            # This means that the gcode contains "\n" as the end-of-line characters and
            # that the gcode should be written out exactly that way.
            newline_handling = ""
            gcode = gcode[2:]
        elif "\r" in gcode:
            # Write out the gcode with whatever end-of-line characters it already has,
            # presumably either "\r" or "\r\n".
            newline_handling = ""
        else:
            # The gcode is assumed to contain "\n" as the end-of-line characters (if
            # there are any end-of-line characters in the gcode).  This case also
            # handles a zero-length gcode string.
            # Write out the gcode but convert "\n" to whatever the system uses.
            # This is also backwards compatible with the "previous" way of doing things.
            newline_handling = None

        if policy.casefold() == "open file dialog":
            dlg = QtGui.QFileDialog()
            dlg.setFileMode(QtGui.QFileDialog.FileMode.AnyFile)
            dlg.setAcceptMode(QtGui.QFileDialog.AcceptMode.AcceptSave)
            dlg.setDirectory(os.path.dirname(filename))
            dlg.selectFile(os.path.basename(filename))
            if dlg.exec_():
                filename = dlg.selectedFiles()[0]
                Path.Log.debug(filename)
                with open(filename, "w", encoding="utf-8", newline=newline_handling) as f:
                    f.write(gcode)
            else:
                return

        elif policy.casefold() == "append unique id on conflict":
            while os.path.isfile(filename):
                base, ext = os.path.splitext(filename)
                filename = f"{base}-1{ext}"
            with open(filename, "w", encoding="utf-8", newline=newline_handling) as f:
                f.write(gcode)

        elif policy.casefold() == "open file dialog on conflict":
            if os.path.isfile(filename):
                dlg = QtGui.QFileDialog()
                dlg.setFileMode(QtGui.QFileDialog.FileMode.AnyFile)
                dlg.setAcceptMode(QtGui.QFileDialog.AcceptSave)
                dlg.setDirectory(os.path.dirname(filename))
                dlg.selectFile(os.path.basename(filename))
                if dlg.exec_():
                    filename = dlg.selectedFiles()[0]
                    Path.Log.debug(filename)
                    with open(filename, "w", encoding="utf-8", newline=newline_handling) as f:
                        f.write(gcode)
                else:
                    return
            else:
                with open(filename, "w", encoding="utf-8", newline=newline_handling) as f:
                    f.write(gcode)

        else:  # Overwrite
            with open(filename, "w", encoding="utf-8", newline=newline_handling) as f:
                f.write(gcode)

        FreeCAD.Console.PrintMessage(f"File written to {filename}\n")

    def Activated(self):
        """
        Handles the activation of post processing, initiating the process based
        on user selection and document context.
        """
        Path.Log.debug(self.candidate.Name)
        FreeCAD.ActiveDocument.openTransaction("Post Process the Selected Job")

        postprocessor_name = _resolve_post_processor_name(self.candidate)
        Path.Log.debug(f"Post Processor: {postprocessor_name}")

        if not postprocessor_name:
            FreeCAD.ActiveDocument.abortTransaction()
            return

        # get a postprocessor
        postprocessor = PostProcessorFactory.get_post_processor(self.candidate, postprocessor_name)

        post_data = postprocessor.export()
        # None is returned if there was an error during argument processing
        # otherwise the "usual" post_data data structure is returned.
        if not post_data:
            FreeCAD.ActiveDocument.abortTransaction()
            return

        policy = Path.Preferences.defaultOutputPolicy()
        generator = FilenameGenerator(job=self.candidate)
        generated_filename = generator.generate_filenames()

        for item in post_data:
            subpart, gcode = item

            # get a name for the file
            subpart = "" if subpart == "allitems" else subpart
            Path.Log.debug(subpart)
            generator.set_subpartname(subpart)
            fname = next(generated_filename)

            #
            # It is useful for a postprocessor to be able to either skip writing out
            # a file or write out a zero-length file to indicate that something unusual
            # has happened.  The "gcode" variable is usually a string containing gcode
            # formatted for output.  If the gcode string is zero length then a zero
            # length file will be written out.  If the "gcode" variable contains None
            # instead, that indicates that the postprocessor doesn't want a file to be
            # written at all.
            #
            # There is at least one old-style postprocessor that currently puts the
            # gcode file out to a file server and doesn't need to write out a file to
            # the system where FreeCAD is running.  In the old-style postprocessors the
            # postprocessor code decided whether to write out a file.  Eventually a
            # newer (more object-oriented) version of that postprocessor will return
            # None for the "gcode" variable value to tell this code not to write out
            # a file.  There may be other uses found for this capability over time.
            #
            if gcode is not None:
                # write the results to the file
                self._write_file(fname, gcode, policy)

        FreeCAD.ActiveDocument.commitTransaction()
        FreeCAD.ActiveDocument.recompute()


if FreeCAD.GuiUp:
    # register the FreeCAD command
    FreeCADGui.addCommand("CAM_Post", CommandPathPost())

FreeCAD.Console.PrintLog("Loading PathPost… done\n")



# ***************************************************************************
# *   Copyright (c) 2023 Uwe Stöhr <uwestoehr@lyx.org>                      *
# *   Copyright (c) 2025 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

__title__ = "FreeCAD FEM constraint current density task panel for the document object"
__author__ = "Uwe Stöhr, Mario Passaglia"
__url__ = "https://www.freecad.org"

## @package task_constraint_currentdensity
#  \ingroup FEM
#  \brief task panel for constraint current density object

from PySide import QtCore

import FreeCAD
import FreeCADGui

from femguiutils import selection_widgets

from femtools import membertools
from . import base_femtaskpanel


class _TaskPanel(base_femtaskpanel._BaseTaskPanel):

    def __init__(self, obj):
        super().__init__(obj)

        self.parameter_widget = FreeCADGui.PySideUic.loadUi(
            FreeCAD.getHomePath() + "Mod/Fem/Resources/ui/CurrentDensity.ui"
        )

        # geometry selection widget
        # start with Solid in list!
        self._selectionWidget = selection_widgets.GeometryElementsSelection(
            obj.References, ["Solid", "Face"], True, False
        )

        # form made from param and selection widget
        self.form = [self.parameter_widget, self._selectionWidget]

        analysis = obj.getParentGroup()
        self._mesh = None
        self._part = None
        if analysis is not None:
            self._mesh = membertools.get_single_member(analysis, "Fem::FemMeshObject")
        if self._mesh is not None:
            self._part = self._mesh.Shape
        self._partVisible = None
        self._meshVisible = None

        QtCore.QObject.connect(
            self.parameter_widget.cb_mode,
            QtCore.SIGNAL("currentIndexChanged(int)"),
            self.mode_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_current_density_1,
            QtCore.SIGNAL("toggled(bool)"),
            self.current_density_1_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_current_density_re_1,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.current_density_re_1_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_current_density_2,
            QtCore.SIGNAL("toggled(bool)"),
            self.current_density_2_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_current_density_re_2,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.current_density_re_2_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.ckb_current_density_3,
            QtCore.SIGNAL("toggled(bool)"),
            self.current_density_3_enabled_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_current_density_re_3,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.current_density_re_3_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_current_density_im_1,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.current_density_im_1_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_current_density_im_2,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.current_density_im_2_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_current_density_im_3,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.current_density_im_3_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_normal_current_density_re,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.normal_current_density_re_changed,
        )
        QtCore.QObject.connect(
            self.parameter_widget.qsb_normal_current_density_im,
            QtCore.SIGNAL("valueChanged(Base::Quantity)"),
            self.normal_current_density_im_changed,
        )

        self.init_parameter_widget()

    def open(self):
        if self._mesh is not None and self._part is not None:
            self._meshVisible = self._mesh.ViewObject.isVisible()
            self._partVisible = self._part.ViewObject.isVisible()
            self._mesh.ViewObject.hide()
            self._part.ViewObject.show()

    def reject(self):
        self._restoreVisibility()
        self._selectionWidget.finish_selection()
        return super().reject()

    def accept(self):
        if self.obj.References != self._selectionWidget.references:
            self.obj.References = self._selectionWidget.references
        self._set_params()
        self._selectionWidget.finish_selection()
        self._restoreVisibility()
        return super().accept()

    def _restoreVisibility(self):
        if self._mesh is not None and self._part is not None:
            if self._meshVisible:
                self._mesh.ViewObject.show()
            else:
                self._mesh.ViewObject.hide()
            if self._partVisible:
                self._part.ViewObject.show()
            else:
                self._part.ViewObject.hide()

    def _get_params(self):
        self.mode = self.obj.Mode

        self.current_density_re_1 = self.obj.CurrentDensity_re_1
        self.current_density_re_2 = self.obj.CurrentDensity_re_2
        self.current_density_re_3 = self.obj.CurrentDensity_re_3
        self.current_density_im_1 = self.obj.CurrentDensity_im_1
        self.current_density_im_2 = self.obj.CurrentDensity_im_2
        self.current_density_im_3 = self.obj.CurrentDensity_im_3

        self.current_density_1_enabled = self.obj.EnableCurrentDensity_1
        self.current_density_2_enabled = self.obj.EnableCurrentDensity_2
        self.current_density_3_enabled = self.obj.EnableCurrentDensity_3

        self.normal_current_density_re = self.obj.NormalCurrentDensity_re
        self.normal_current_density_im = self.obj.NormalCurrentDensity_im

    def _set_params(self):
        self.obj.Mode = self.mode

        self.obj.CurrentDensity_re_1 = self.current_density_re_1
        self.obj.CurrentDensity_re_2 = self.current_density_re_2
        self.obj.CurrentDensity_re_3 = self.current_density_re_3
        self.obj.CurrentDensity_im_1 = self.current_density_im_1
        self.obj.CurrentDensity_im_2 = self.current_density_im_2
        self.obj.CurrentDensity_im_3 = self.current_density_im_3

        self.obj.EnableCurrentDensity_1 = self.current_density_1_enabled
        self.obj.EnableCurrentDensity_2 = self.current_density_2_enabled
        self.obj.EnableCurrentDensity_3 = self.current_density_3_enabled

        self.obj.NormalCurrentDensity_re = self.normal_current_density_re
        self.obj.NormalCurrentDensity_im = self.normal_current_density_im

    def init_parameter_widget(self):
        self._get_params()

        # custom current density
        self.parameter_widget.qsb_current_density_re_1.setProperty(
            "value", self.current_density_re_1
        )
        self.parameter_widget.qsb_current_density_re_1.setEnabled(self.current_density_1_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_current_density_re_1).bind(
            self.obj, "CurrentDensity_re_1"
        )

        self.parameter_widget.qsb_current_density_re_2.setProperty(
            "value", self.current_density_re_2
        )
        self.parameter_widget.qsb_current_density_re_2.setEnabled(self.current_density_2_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_current_density_re_2).bind(
            self.obj, "CurrentDensity_re_2"
        )

        self.parameter_widget.qsb_current_density_re_3.setProperty(
            "value", self.current_density_re_3
        )
        self.parameter_widget.qsb_current_density_re_3.setEnabled(self.current_density_3_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_current_density_re_3).bind(
            self.obj, "CurrentDensity_re_3"
        )

        self.parameter_widget.qsb_current_density_im_1.setProperty(
            "value", self.current_density_im_1
        )
        self.parameter_widget.qsb_current_density_im_1.setEnabled(self.current_density_1_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_current_density_im_1).bind(
            self.obj, "CurrentDensity_im_1"
        )

        self.parameter_widget.qsb_current_density_im_2.setProperty(
            "value", self.current_density_im_2
        )
        self.parameter_widget.qsb_current_density_im_2.setEnabled(self.current_density_2_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_current_density_im_2).bind(
            self.obj, "CurrentDensity_im_2"
        )

        self.parameter_widget.qsb_current_density_im_3.setProperty(
            "value", self.current_density_im_3
        )
        self.parameter_widget.qsb_current_density_im_3.setEnabled(self.current_density_3_enabled)
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_current_density_im_3).bind(
            self.obj, "CurrentDensity_im_3"
        )

        self.parameter_widget.ckb_current_density_1.setChecked(self.current_density_1_enabled)
        self.parameter_widget.ckb_current_density_2.setChecked(self.current_density_2_enabled)
        self.parameter_widget.ckb_current_density_3.setChecked(self.current_density_3_enabled)

        self.parameter_widget.qsb_normal_current_density_re.setProperty(
            "value", self.normal_current_density_re
        )
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_normal_current_density_re).bind(
            self.obj, "NormalCurrentDensity_re"
        )
        self.parameter_widget.qsb_normal_current_density_im.setProperty(
            "value", self.normal_current_density_im
        )
        FreeCADGui.ExpressionBinding(self.parameter_widget.qsb_normal_current_density_im).bind(
            self.obj, "NormalCurrentDensity_im"
        )

        self.mode_enum = self.obj.getEnumerationsOfProperty("Mode")
        index = self.mode_enum.index(self.mode)
        self.parameter_widget.cb_mode.addItems(self.mode_enum)
        self.parameter_widget.cb_mode.setCurrentIndex(index)

    def current_density_1_enabled_changed(self, value):
        self.current_density_1_enabled = value
        self.parameter_widget.qsb_current_density_re_1.setEnabled(value)
        self.parameter_widget.qsb_current_density_im_1.setEnabled(value)

    def current_density_2_enabled_changed(self, value):
        self.current_density_2_enabled = value
        self.parameter_widget.qsb_current_density_re_2.setEnabled(value)
        self.parameter_widget.qsb_current_density_im_2.setEnabled(value)

    def current_density_3_enabled_changed(self, value):
        self.current_density_3_enabled = value
        self.parameter_widget.qsb_current_density_re_3.setEnabled(value)
        self.parameter_widget.qsb_current_density_im_3.setEnabled(value)

    def current_density_re_1_changed(self, value):
        self.current_density_re_1 = value

    def current_density_re_2_changed(self, value):
        self.current_density_re_2 = value

    def current_density_re_3_changed(self, value):
        self.current_density_re_3 = value

    def current_density_im_1_changed(self, value):
        self.current_density_im_1 = value

    def current_density_im_2_changed(self, value):
        self.current_density_im_2 = value

    def current_density_im_3_changed(self, value):
        self.current_density_im_3 = value

    def normal_current_density_re_changed(self, value):
        self.normal_current_density_re = value

    def normal_current_density_im_changed(self, value):
        self.normal_current_density_im = value

    def mode_changed(self, index):
        self.mode = self.mode_enum[index]
        if self.mode == "Custom":
            self.parameter_widget.gb_custom.setEnabled(True)
            self.parameter_widget.gb_normal.setEnabled(False)
        elif self.mode == "Normal":
            self.parameter_widget.gb_custom.setEnabled(False)
            self.parameter_widget.gb_normal.setEnabled(True)



#!/usr/bin/env python
# ***************************************************************************
# *   Copyright (c) 2013 WandererFan <wandererfan@gmail.com>                *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************/

# Tester for Draft makePathArray - shapes on a path - without subelements (see testPathArraySel.py)
# Usage: in FC gui, select a "shape" document object (sphere, box, etc) (!!select in
# tree, not document view!!), then a "wire" document object (Wire, Circle, Rectangle,
# DWire, etc) then run this macro.

import FreeCAD
import FreeCADGui
import Part
import Draft

print("testPathArray started")
items = 4  # count
centretrans = FreeCAD.Vector(0, 0, 0)  # no translation
# centretrans = FreeCAD.Vector(-5,-5,0)                     # translation
orient = True  # align to curve
# orient = False                                            # don't align to curve

s = FreeCADGui.Selection.getSelection()
print("testPathArray: Objects in selection: ", len(s))
print("First object in selection is a: ", s[0].Shape.ShapeType)
print("Second object in selection is a: ", s[1].Shape.ShapeType)
base = s[0]
path = s[1]
pathsubs = []

# o = Draft.makePathArray(base,path,items)                            # test with defaults
o = Draft.makePathArray(base, path, items, centretrans, orient, pathsubs)  # test with non-defaults

print("testPathArray ended")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


__title__ = "FreeCAD material card importer"
__author__ = "Juergen Riegel"
__url__ = "https://www.freecad.org"


import os
import glob
import uuid
from pathlib import Path
import xml.etree.ElementTree as ET

def decode(name):
    "decodes encoded strings"
    try:
        decodedName = (name.decode("utf8"))
    except UnicodeDecodeError:
        try:
            decodedName = (name.decode("latin1"))
        except UnicodeDecodeError:
            print("Error: Couldn't determine character encoding")
            decodedName = name
    return decodedName

def read(filename):
    "reads a FCMat file and returns a dictionary from it"

    # the reader returns a dictionary in any case even if the file  has problems
    # an empty dict is returned in such case

    card_name_file = os.path.splitext(os.path.basename(filename))[0]
    f = open(filename, encoding="utf8")
    try:
        content = f.readlines()
    except Exception:
        # https://forum.freecad.org/viewtopic.php?f=18&t=56912#p489721
        # older FreeCAD do not write utf-8 for special character on windows
        # I have seen "ISO-8859-15" or "windows-1252"
        # explicit utf-8 writing, https://github.com/FreeCAD/FreeCAD/commit/9a564dd906f
        print("Error on card loading. File might not utf-8.")
        error_message = "Error on loading. Material file '{}' might not utf-8.".format(filename)
        print("{}\n".format(error_message))
        return {}
    d = {}
    d["Meta"] = {}
    d["General"] = {}
    d["Mechanical"] = {}
    d["Fluidic"] = {}
    d["Thermal"] = {}
    d["Electromagnetic"] = {}
    d["Architectural"] = {}
    d["Rendering"] = {}
    d["VectorRendering"] = {}
    d["Cost"] = {}
    d["UserDefined"] = {}
    d["Meta"]["CardName"] = card_name_file  # CardName is the MatCard file name
    section = ''
    for ln, line in enumerate(content):
        # line numbers are used for CardName and AuthorAndLicense
        # the use of line number is not smart for a data model
        # a wrong user edit could break the file

        # comment
        if line.startswith('#'):
            # a '#' is assumed to be a comment which is ignored
            continue
        # CardName
        if line.startswith(';') and ln == 0:
            pass

        # AuthorAndLicense
        elif line.startswith(';') and ln == 1:
            v = line.split(";")[1].strip()  # Line 2
            if hasattr(v, "decode"):
                v = v.decode('utf-8')
            d["General"]["AuthorAndLicense"] = v # Move the field to the general group

        # rest
        else:
            # ; is a Comment
            # [ is a Section
            if line[0] == '[':
                line = line[1:]
                k = line.split("]", 1)
                if len(k) >= 2:
                    v = k[0].strip()
                    if hasattr(v, "decode"):
                        v = v.decode('utf-8')
                    section = v
            elif line[0] not in ";":
                # split once on first occurrence
                # a link could contain a '=' and thus would be split
                k = line.split("=", 1)
                if len(k) == 2:
                    v = k[1].strip()
                    if hasattr(v, "decode"):
                        v = v.decode('utf-8')
                    d[section][k[0].strip()] = v
    return d

def yamGeneral(card):
    father = ""
    materialStandard = ""
    yamModels = ""
    yam = "# File created by ConvertFCMat.py\n"
    yam += "General:\n"

    # Add UUIDs
    yam += '  UUID: "{0}"\n'.format(uuid.uuid4())
    for param in card:
        if param in ["Name", "AuthorAndLicense", "Description", "ReferenceSource", "SourceURL"]:
            yam += '  {0}: "{1}"\n'.format(param, card[param])
        elif param  in ["Father"]:
            father += '    {0}: "{1}"\n'.format(param, card[param])
        elif param  in ["KindOfMaterial", "MaterialNumber", "Norm", "StandardCode"]:
            if param == "Norm": # Handle the name change
                materialStandard += '    {0}: "{1}"\n'.format("StandardCode", card[param])
            else:
                materialStandard += '    {0}: "{1}"\n'.format(param, card[param])

    if len(father) > 0:
        yamModels += "  {0}:\n".format('Father')
        yamModels += "    UUID: '{0}'\n".format('9cdda8b6-b606-4778-8f13-3934d8668e67')
        yamModels += father
    if len(materialStandard) > 0:
        yamModels += "  {0}:\n".format('MaterialStandard')
        yamModels += "    UUID: '{0}'\n".format('1e2c0088-904a-4537-925f-64064c07d700')
        yamModels += materialStandard

    return yam, yamModels

def yamSection(card, header, uuid):
    if len(card) > 0:
        yam = "  {0}:\n".format(header)
        yam += "    UUID: '{0}'\n".format(uuid)
        for param in card:
            yam += '    {0}: "{1}"\n'.format(param, card[param])
    else:
        yam = ""

    return yam

def yamMechanical(card):
    # Check which model we need
    useDensity = False
    useIso = False
    useLinearElastic = False
    for param in card:
        if param in ["Density"]:
            useDensity = True
        elif param in ["BulkModulus", "PoissonRatio", "ShearModulus", "YoungsModulus"]:
            useIso = True
        elif param in ["AngleOfFriction", "CompressiveStrength", "FractureToughness",
                       "UltimateStrain", "UltimateTensileStrength", "YieldStrength", "Stiffness", "Hardness"]:
            useLinearElastic = True

    yam = ""
    if useLinearElastic:
        return yamSection(card, 'LinearElastic', '7b561d1d-fb9b-44f6-9da9-56a4f74d7536')
    if useIso:
        yam = yamSection(card, 'IsotropicLinearElastic', 'f6f9e48c-b116-4e82-ad7f-3659a9219c50')
    if useDensity:
        return yam + yamSection(card, 'Density', '454661e5-265b-4320-8e6f-fcf6223ac3af')

    # default mechanical model
    return ""

def yamFluid(card):
    # Split out density
    for param in card:
        if param not in ["Density"]:
            return yamSection(card, 'Fluid', '1ae66d8c-1ba1-4211-ad12-b9917573b202')

    return yamSection(card, 'Density', '454661e5-265b-4320-8e6f-fcf6223ac3af')

def yamThermal(card):
    return yamSection(card, 'Thermal', '9959d007-a970-4ea7-bae4-3eb1b8b883c7')

def yamElectromagnetic(card):
    return yamSection(card, 'Electromagnetic', 'b2eb5f48-74b3-4193-9fbb-948674f427f3')

def yamArchitectural(card):
    return yamSection(card, 'Architectural', '32439c3b-262f-4b7b-99a8-f7f44e5894c8')

def yamCost(card):
    return yamSection(card, 'Costs', '881df808-8726-4c2e-be38-688bb6cce466')

def yamRendering(card):
    # Check which model we need
    useTexture = False
    useAdvanced = False
    for param in card:
        if param in ["TexturePath", "TextureScaling"]:
            useTexture = True
        elif param in ["FragmentShader", "VertexShader"]:
            useAdvanced = True

    if useAdvanced:
        return yamSection(card, 'AdvancedRendering', 'c880f092-cdae-43d6-a24b-55e884aacbbf')
    if useTexture:
        return yamSection(card, 'TextureRendering', 'bbdcc65b-67ca-489c-bd5c-a36e33d1c160')

    # default rendering model
    return yamSection(card, 'BasicRendering', 'f006c7e4-35b7-43d5-bbf9-c5d572309e6e')

def yamVectorRendering(card):
    return yamSection(card, 'VectorRendering', 'fdf5a80e-de50-4157-b2e5-b6e5f88b680e')

def saveYaml(card, output):
    yam, yamModels = yamGeneral(card["General"])
    if len(card["Mechanical"]) > 0 or \
        len(card["Fluidic"]) > 0 or \
        len(card["Thermal"]) > 0 or \
        len(card["Electromagnetic"]) > 0 or \
        len(card["Architectural"]) > 0 or \
        len(card["Cost"]) > 0 or \
        len(yamModels) > 0:
        yam += "Models:\n"
        yam += yamModels
        if "Mechanical" in card:
            yam += yamMechanical(card["Mechanical"])
        if "Fluidic" in card:
            yam += yamFluid(card["Fluidic"])
        if "Thermal" in card:
            yam += yamThermal(card["Thermal"])
        if "Electromagnetic" in card:
            yam += yamElectromagnetic(card["Electromagnetic"])
        if "Architectural" in card:
            yam += yamArchitectural(card["Architectural"])
        if "Cost" in card:
            yam += yamCost(card["Cost"])
    if len(card["Rendering"]) > 0 or len(card["VectorRendering"]) > 0:
        yam += "AppearanceModels:\n"
        if "Rendering" in card:
            yam += yamRendering(card["Rendering"])
        if "VectorRendering" in card:
            yam += yamVectorRendering(card["VectorRendering"])

    file = open(output, "w", encoding="utf-8")
    file.write(yam)
    file.close()

def convert(infolder, outfolder):
    a_path = infolder + '/**/*.FCMat'
    dir_path_list = glob.glob(a_path, recursive=True)

    for a_path in dir_path_list:
        p = Path(a_path)
        relative = p.relative_to(infolder)
        out = Path(outfolder) / relative
        print("('{0}', '{1}') -> {2}".format(infolder, relative, out))

        try:
            card = read(p)
        except Exception:
            print("Error converting card '{0}'. Skipped.")
            continue

        out.parent.mkdir(parents=True, exist_ok=True)
        saveYaml(card, out)

import argparse
parser = argparse.ArgumentParser()
parser.add_argument("infolder", help="Input folder containing older material cards")
parser.add_argument("outfolder", help="Output folder to place the converted material cards")
args = parser.parse_args()

print("Input folder '{0}'".format(args.infolder))
print("Output folder '{0}'".format(args.outfolder))

convert(args.infolder, args.outfolder)


# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides functions to repair certain objects created with old versions."""
## @package heal
# \ingroup draftfunctions
# \brief Provides functions to repair certain objects from old versions.

## \addtogroup draftfunctions
# @{
import FreeCAD as App
import draftutils.utils as utils

from draftmake.make_copy import make_copy


def heal(objlist=None, delete=True, reparent=True):
    """heal([objlist],[delete],[reparent])

    Recreate Draft objects that are damaged, for example if created from an
    earlier version. If ran without arguments, all the objects in the document
    will be healed if they are damaged.

    Parameters
    ----------
    objlist : list

    delete : Base.Vector or list of Base.Vector
        If delete is True, the damaged objects are deleted (default).

    reparent : bool
        If reparent is True (default), new objects go at the very same place
        in the tree than their original.
    """

    auto = False

    if not objlist:
        objlist = App.ActiveDocument.Objects
        print("Automatic mode: Healing whole document…")
        auto = True
    else:
        print("Manual mode: Force-healing selected objects…")

    if not isinstance(objlist, list):
        objlist = [objlist]

    dellist = []
    got = False

    for obj in objlist:
        dtype = utils.get_type(obj)
        ftype = obj.TypeId
        if ftype in ["Part::FeaturePython", "App::FeaturePython", "Part::Part2DObjectPython"]:
            proxy = obj.Proxy
            if hasattr(obj, "ViewObject"):
                if hasattr(obj.ViewObject, "Proxy"):
                    proxy = obj.ViewObject.Proxy
            if (proxy == 1) or (dtype in ["Unknown", "Part"]) or (not auto):
                got = True
                dellist.append(obj.Name)
                props = obj.PropertiesList
                if ("Dimline" in props) and ("Start" in props):
                    print("Healing " + obj.Name + " of type Dimension")
                    nobj = make_copy(obj, force="Dimension", reparent=reparent)
                elif ("Height" in props) and ("Length" in props):
                    print("Healing " + obj.Name + " of type Rectangle")
                    nobj = make_copy(obj, force="Rectangle", reparent=reparent)
                elif ("Points" in props) and ("Closed" in props):
                    if "BSpline" in obj.Name:
                        print("Healing " + obj.Name + " of type BSpline")
                        nobj = make_copy(obj, force="BSpline", reparent=reparent)
                    else:
                        print("Healing " + obj.Name + " of type Wire")
                        nobj = make_copy(obj, force="Wire", reparent=reparent)
                elif ("Radius" in props) and ("FirstAngle" in props):
                    print("Healing " + obj.Name + " of type Circle")
                    nobj = make_copy(obj, force="Circle", reparent=reparent)
                elif ("DrawMode" in props) and ("FacesNumber" in props):
                    print("Healing " + obj.Name + " of type Polygon")
                    nobj = make_copy(obj, force="Polygon", reparent=reparent)
                else:
                    dellist.pop()
                    print("Object " + obj.Name + " is not healable")

    if not got:
        print("No object seems to need healing")
    else:
        print("Healed ", len(dellist), " objects")

    if dellist and delete:
        for n in dellist:
            App.ActiveDocument.removeObject(n)


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

from __future__ import annotations
from dataclasses import dataclass, field
from enum import Enum
from typing import List, Optional


#
# Enums
#


class ParameterType(str, Enum):
    BOOLEAN = "Boolean"
    INT = "Int"
    LONG = "Long"
    STRING = "String"
    OBJECT = "Object"
    FLOAT = "Float"
    COMPLEX = "Complex"
    CHAR = "Char"
    TUPLE = "Tuple"
    LIST = "List"
    DICT = "Dict"
    MODULE = "Module"
    CALLABLE = "Callable"
    SEQUENCE = "Sequence"
    VECTOR = "Vector"

    def __str__(self):
        return self.value


#
# Supporting Classes
#


@dataclass
class Author:
    """Represents the <Author> element inside <Documentation>."""

    # The text content of <Author> is effectively a string;
    # we capture it in `content` to hold the text node if needed.
    content: Optional[str] = None

    # Attributes
    Name: str = "FreeCAD Project"
    EMail: str = "example@freecad.org"
    Licence: str = "LGPL"


@dataclass
class Documentation:
    """
    Corresponds to the <Documentation> element.
    Can contain an <Author>, <DeveloperDocu>, and <UserDocu>.
    """

    Author: Optional[Author] = None
    DeveloperDocu: Optional[str] = None
    UserDocu: Optional[str] = None


@dataclass
class Property:
    """
    Corresponds to <Property> in the schema.
    It has required attributes Name, Type and optional StartValue,
    plus optional child <Documentation>.
    """

    # Child
    Documentation: Optional[Documentation] = None

    # Attributes
    Name: str = ""
    Type: str = ""
    StartValue: Optional[str] = None


@dataclass
class ViewProvider:
    """
    Corresponds to <ViewProvider>, which can contain 0..∞ <Property> children.
    """

    Property: List[Property] = field(default_factory=list)


@dataclass
class Parameter:
    """
    Corresponds to <Parameter> in the schema.
    It has a required 'Name' (str) and a required 'Type' (enumeration).
    """

    Name: str
    Type: ParameterType


#
# Elements under <PythonExport>
#


@dataclass
class Methode:
    """
    Corresponds to <Methode> inside <PythonExport>.
    Contains an optional <Documentation> and 0..∞ <Parameter>.
    """

    Documentation: Optional[Documentation] = None
    Parameter: List[Parameter] = field(default_factory=list)

    # Attributes
    Name: str = ""
    Const: Optional[bool] = None
    Keyword: bool = False
    NoArgs: bool = False
    Class: bool = False
    Static: bool = False


@dataclass
class Attribute:
    """
    Corresponds to <Attribute> inside <PythonExport>.
    It has a required <Documentation>, a required <Parameter>,
    and attributes Name, ReadOnly.
    """

    Documentation: Documentation
    Parameter: Parameter

    # Attributes
    Name: str
    ReadOnly: bool


@dataclass
class SequenceProtocol:
    """
    Corresponds to the <Sequence> element inside <PythonExport>.
    All attributes are required booleans.
    """

    sq_length: bool = False
    sq_concat: bool = False
    sq_repeat: bool = False
    sq_item: bool = False
    mp_subscript: bool = False
    sq_ass_item: bool = False
    mp_ass_subscript: bool = False
    sq_contains: bool = False
    sq_inplace_concat: bool = False
    sq_inplace_repeat: bool = False


@dataclass
class PythonExport:
    """
    Corresponds to <PythonExport> inside <GenerateModel>.
    It contains:
      - optional <Documentation>
      - 0..∞ <Methode>
      - 0..∞ <Attribute>
      - optional <Sequence>
      - optional <CustomAttributes>
      - one <ClassDeclarations> (type=string)
      - one <ForwardDeclarations> (type=string)
    Plus many attributes with required/optional flags.
    """

    Documentation: Optional[Documentation] = None
    Methode: List[Methode] = field(default_factory=list)
    Attribute: List[Attribute] = field(default_factory=list)
    Sequence: Optional[SequenceProtocol] = None
    CustomAttributes: Optional[str] = ""  # To match the original XML model
    ClassDeclarations: str = ""
    ForwardDeclarations: str = ""
    NoArgs: bool = False

    # Attributes
    ModuleName: str = ""
    Name: str = ""
    PythonName: Optional[str] = None
    Include: str = ""
    Father: str = ""
    Twin: str = ""
    Namespace: str = ""
    FatherInclude: str = ""
    FatherNamespace: str = ""
    Constructor: bool = False
    NumberProtocol: bool = False
    RichCompare: bool = False
    TwinPointer: str = ""
    Delete: bool = False
    Reference: Optional[bool] = None
    Initialization: bool = False
    DisableNotify: bool = False
    DescriptorGetter: bool = False
    DescriptorSetter: bool = False
    IsExplicitlyExported: bool = False


#
# Module-Related Classes
#


@dataclass
class Dependencies:
    """
    Corresponds to the <Dependencies> element inside <Module>.
    It contains 0..∞ local <Module> elements which are not typed in the XSD.
    We'll treat these as strings or possibly minimal structures.
    """

    Module: List[str] = field(default_factory=list)


@dataclass
class Feature:
    """
    Corresponds to <Feature> in <Module>'s <Content>.
    Has optional <Documentation>, 0..∞ <Property>, optional <ViewProvider>,
    and a required attribute 'Name'.
    """

    Documentation: Optional[Documentation] = None
    Property: List[Property] = field(default_factory=list)
    ViewProvider: Optional[ViewProvider] = None

    # Attributes
    Name: str = ""


@dataclass
class DocObject:
    """
    Corresponds to <DocObject> in <Module>'s <Content>.
    Has optional <Documentation>, 0..∞ <Property>, and a required 'Name' attribute.
    """

    Documentation: Optional[Documentation] = None
    Property: List[Property] = field(default_factory=list)

    # Attributes
    Name: str = ""


@dataclass
class ModuleContent:
    """
    Corresponds to the <Content> element in <Module>.
    Contains:
      - 0..∞ <Property>
      - 0..∞ <Feature>
      - 0..∞ <DocObject>
      - 0..∞ <GuiCommand>
      - 0..∞ <PreferencesPage>
    """

    Property: List[Property] = field(default_factory=list)
    Feature: List[Feature] = field(default_factory=list)
    DocObject: List[DocObject] = field(default_factory=list)
    GuiCommand: List[str] = field(default_factory=list)
    PreferencesPage: List[str] = field(default_factory=list)


@dataclass
class Module:
    """
    Corresponds to the top-level <Module> element.
    Has optional <Documentation>, optional <Dependencies>,
    a required <Content>, and a required attribute Name.
    """

    Documentation: Optional[Documentation] = None
    Dependencies: Optional[Dependencies] = None
    Content: ModuleContent = field(default_factory=ModuleContent)

    # Attributes
    Name: str = ""


#
# Root Element
#


@dataclass
class GenerateModel:
    """
    Corresponds to the root element <GenerateModel>.
    Contains 0..∞ <Module> and 0..∞ <PythonExport>.
    """

    Module: List[Module] = field(default_factory=list)
    PythonExport: List[PythonExport] = field(default_factory=list)

    def dump(self):
        # Print or process the resulting GenerateModel object
        print("Parsed GenerateModel object:")

        if self.PythonExport:
            py_exp = self.PythonExport[0]
            print("PythonExport Name:", py_exp.Name)
            if py_exp.Documentation and py_exp.Documentation.Author:
                print("Author Name:", py_exp.Documentation.Author.Name)
                print("Author Email:", py_exp.Documentation.Author.EMail)
                print("Author Licence:", py_exp.Documentation.Author.Licence)
            print("DeveloperDocu:", py_exp.Documentation.DeveloperDocu)
            print("UserDocu:", py_exp.Documentation.UserDocu)

            print("Class Attributes:")
            for attr in py_exp.Attribute:
                print(f"  - {attr.Name} (type={attr.Parameter.Type}, readOnly={attr.ReadOnly})")

            print("Methods:")
            for meth in py_exp.Methode:
                print(f"  - {meth.Name}")
                # Each method might have parameters
                for param in meth.Parameter:
                    print(f"    * param: {param.Name}, type={param.Type}")



from typing import Any, Final

from Base.BaseClass import BaseClass
from Base.Metadata import export

@export(
    FatherInclude="Base/BaseClassPy.h",
    Include="Mod/CAM/PathSimulator/App/PathSim.h",
    Namespace="PathSimulator",
    Constructor=True,
    Delete=True,
)
class PathSim(BaseClass):
    """
    FreeCAD python wrapper of PathSimulator

    PathSimulator.PathSim():

    Create a path simulator object

    Author: Shai Seger (shaise_at_g-mail)
    License: LGPL-2.1-or-later
    """

    def BeginSimulation(self, **kwargs) -> Any:
        """BeginSimulation(stock, resolution):

        Start a simulation process on a box shape stock with given resolution"""
        ...

    def SetToolShape(self) -> Any:
        """SetToolShape(shape):

        Set the shape of the tool to be used for simulation"""
        ...

    def GetResultMesh(self) -> Any:
        """
        GetResultMesh():

                  Return the current mesh result of the simulation."""
        ...

    def ApplyCommand(self, **kwargs) -> Any:
        """
        ApplyCommand(placement, command):

                  Apply a single path command on the stock starting from placement."""
        ...
    Tool: Final[Any]
    """Return current simulation tool."""



# SPDX-License-Identifier: LGPL-2.1-or-later

import FreeCAD
from FreeCAD import Vector, Base, newDocument, closeDocument
import Part

import math
import os

if "BUILD_SKETCHER" in FreeCAD.__cmake__:
    import Sketcher

import unittest


class RegressionTests(unittest.TestCase):

    # pylint: disable=attribute-defined-outside-init

    def setUp(self):
        """Create a document for each test in the test suite"""
        self.Doc = newDocument("PartRegressionTest." + self._testMethodName)
        self.KeepTestDoc = False

    def test_issue_4456(self):
        """
        0004456: Regression : Part.Plane.Intersect does not accept plane as argument
        """
        p1 = Part.Plane()
        p2 = Part.Plane(Vector(0, 0, 0), Vector(1, 0, 0))

        result = p1.intersect(p2)
        line = result.pop()
        self.assertEqual(line.Location, Vector(0, 0, 0))
        self.assertEqual(line.Direction, Vector(0, 1, 0))
        # We should now have empty list...
        with self.assertRaises(IndexError):
            result.pop()

    def test_issue_15735(self):
        """
        15735: Point in sketch as loft profile won't work in dev, but works in stable
        The following test is a simplified version of the issue, but the outcome is the same
        """

        if "BUILD_SKETCHER" in FreeCAD.__cmake__:
            # Arrange
            ArcSketch = self.Doc.addObject("Sketcher::SketchObject", "ArcSketch")
            ArcSketch.Placement = Base.Placement(
                Base.Vector(0.000000, 0.000000, 0.000000),
                Base.Rotation(0.500000, 0.500000, 0.500000, 0.500000),
            )
            ArcSketch.MapMode = "Deactivated"

            geoList = []
            geoList.append(
                Part.ArcOfCircle(
                    Part.Circle(
                        Base.Vector(0.000000, 0.000000, 0.000000),
                        Base.Vector(0.000000, 0.000000, 1.000000),
                        10.000000,
                    ),
                    3.141593,
                    6.283185,
                )
            )
            ArcSketch.addGeometry(geoList, False)
            del geoList

            constraintList = []
            ArcSketch.addConstraint(Sketcher.Constraint("Radius", 0, 10.000000))
            constraintList.append(Sketcher.Constraint("Coincident", 0, 3, -1, 1))
            constraintList.append(Sketcher.Constraint("PointOnObject", 0, 2, -1))
            constraintList.append(Sketcher.Constraint("PointOnObject", 0, 1, -1))
            ArcSketch.addConstraint(constraintList)
            del constraintList

            self.Doc.recompute()

            PointSketch = self.Doc.addObject("Sketcher::SketchObject", "PointSketch")
            PointSketch.Placement = Base.Placement(
                Base.Vector(-10.000000, 0.000000, 0.000000),
                Base.Rotation(0.500000, 0.500000, 0.500000, 0.500000),
            )
            PointSketch.MapMode = "Deactivated"

            PointSketch.addGeometry(Part.Point(Base.Vector(0.000000, 0.000000, 0)))

            PointSketch.addConstraint(Sketcher.Constraint("Coincident", 0, 1, -1, 1))

            self.Doc.recompute()

            Loft = self.Doc.addObject("Part::Loft", "Loft")
            Loft.Sections = [
                ArcSketch,
                PointSketch,
            ]
            Loft.Solid = False
            Loft.Ruled = False
            Loft.Closed = False

            # Act
            self.Doc.recompute()

            # Assert
            self.assertTrue(Loft.isValid())
            self.KeepTestDoc = not Loft.isValid()

    def test_OptimalBox(self):
        box = Part.makeBox(1, 1, 1)
        self.assertTrue(box.optimalBoundingBox(True, False).isValid())

    def test_CircularReference(self):

        cube = self.Doc.addObject("Part::Box", "Cube")
        cube.setExpression("Length", "Width + 10mm")
        with self.assertRaises(RuntimeError) as context:
            cube.setExpression("Width", "Length + 10mm")
        assert "Width reference creates a cyclic dependency." in str(context.exception)

        cube.setExpression(".Placement.Base.x", ".Placement.Base.y + 10mm")
        with self.assertRaises(RuntimeError) as context:
            cube.setExpression(".Placement.Base.y", ".Placement.Base.x + 10mm")
        assert ".Placement.Base.y reference creates a cyclic dependency." in str(context.exception)

        cube.recompute()
        v1 = cube.Placement.Base
        cube.recompute()
        assert cube.Placement.Base.isEqual(v1, 1e-6)

    def test_TangentMode3_Issue24254(self):
        """In FreeCAD 1.1 we changed the behavior of the mmTangentPlane, but added code to handle old files
        that relied upon the original tangent plane behavior. This test ensures that old files open with the expected
        tangent plane setup."""

        location = os.path.dirname(os.path.realpath(__file__))
        FreeCAD.openDocument(os.path.join(location, "TestTangentMode3-0.21.FCStd"), True)

        def check_plane(name, expected_pos, expected_normal, expected_xaxis):
            obj = FreeCAD.ActiveDocument.getObject(name)
            if not obj:
                FreeCAD.Console.PrintError(f"{name}: object not found\n")
                return

            pos = obj.Placement.Base
            rot = obj.Placement.Rotation
            normal = rot.multVec(FreeCAD.Vector(0, 0, 1))
            xaxis = rot.multVec(FreeCAD.Vector(1, 0, 0))

            # position
            self.assertAlmostEqual((pos - expected_pos).Length, 0)

            # normal
            self.assertAlmostEqual(normal.getAngle(expected_normal), 0)

            # tangent x-axis
            self.assertAlmostEqual(xaxis.getAngle(expected_xaxis), 0)

        r = 10.0
        rad30 = math.radians(30)
        cos30 = math.cos(rad30)
        sin30 = math.sin(rad30)

        expected_planes = [
            (
                "Sketch001",
                FreeCAD.Vector(-1 * r, 0, 0),  # position
                FreeCAD.Vector(1, 0, 0),  # normal
                FreeCAD.Vector(0, -1, 0),
            ),  # tangent x-axis
            (
                "Sketch002",
                FreeCAD.Vector(sin30 * r, cos30 * cos30 * (-r), cos30 * sin30 * r),
                FreeCAD.Vector(sin30, cos30 * cos30 * (-1), cos30 * sin30),
                FreeCAD.Vector(0, sin30, cos30),
            ),
            (
                "Sketch003",
                FreeCAD.Vector(cos30 * cos30 * r, sin30 * r, cos30 * sin30 * r),
                FreeCAD.Vector(cos30 * cos30, sin30, cos30 * sin30),
                FreeCAD.Vector(sin30, 0, -cos30),
            ),
        ]

        for name, pos, normal, xaxis in expected_planes:
            check_plane(name, pos, normal, xaxis)

    def tearDown(self):
        """Clean up our test, optionally preserving the test document"""
        # This flag allows doing something like this:
        #   self.KeepTestDoc = True
        #   import TestApp
        #   TestApp.Test("TestPartApp.RegressionTests.test_issue_15735")
        # to leave the test document(s) around for further examination in an interactive setting.
        if hasattr(self, "KeepTestDoc") and self.KeepTestDoc:
            return
        closeDocument(self.Doc.Name)



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2017 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""The BIM Rewire command"""

import FreeCAD
import FreeCADGui


QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP


class BIM_Rewire:

    def GetResources(self):
        return {
            "Pixmap": "BIM_Rewire",
            "MenuText": QT_TRANSLATE_NOOP("BIM_Rewire", "Rewire"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_Rewire", "Recreates wires from selected objects"),
            "Accel": "R,W",
        }

    def IsActive(self):
        v = hasattr(FreeCADGui.getMainWindow().getActiveWindow(), "getSceneGraph")
        return v

    def Activated(self):
        import Part
        import Draft
        import DraftGeomUtils

        objs = FreeCADGui.Selection.getSelection()
        names = []
        edges = []
        for obj in objs:
            if hasattr(obj, "Shape") and hasattr(obj.Shape, "Edges") and obj.Shape.Edges:
                edges.extend(obj.Shape.Edges)
                names.append(obj.Name)
        wires = DraftGeomUtils.findWires(edges)
        FreeCAD.ActiveDocument.openTransaction("Rewire")
        selectlist = []
        for wire in wires:
            if DraftGeomUtils.hasCurves(wire):
                nobj = FreeCAD.ActiveDocument.addObject("Part::Feature", "Wire")
                nobj.shape = wire
                selectlist.append(nobj)
            else:
                selectlist.append(Draft.makeWire([v.Point for v in wire.OrderedVertexes]))
        for name in names:
            FreeCAD.ActiveDocument.removeObject(name)
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.Selection.clearSelection()
        for obj in selectlist:
            FreeCADGui.Selection.addSelection(obj)
        FreeCAD.ActiveDocument.recompute()


FreeCADGui.addCommand("BIM_Rewire", BIM_Rewire())



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2024 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

"""This contains nativeifc status widgets and functionality"""

import csv
import os

import FreeCAD
import FreeCADGui

translate = FreeCAD.Qt.translate
PARAMS = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/NativeIFC")
text_on = translate("BIM", "Strict IFC mode is ON (all objects are IFC)")
text_off = translate("BIM", "Strict IFC mode is OFF (IFC and non-IFC objects allowed)")


def set_status_widget(statuswidget):
    """Adds the needed controls to the status bar"""

    from PySide import QtGui  # lazy import
    import Arch_rc

    # lock button
    lock_button = QtGui.QAction()
    icon = QtGui.QIcon(":/icons/IFC.svg")
    lock_button.setIcon(icon)
    lock_button.setCheckable(True)
    doc = FreeCAD.ActiveDocument
    statuswidget.addAction(lock_button)
    statuswidget.lock_button = lock_button
    if doc and "IfcFilePath" in doc.PropertiesList:
        checked = True
    else:
        checked = False
    # set the button first, without converting the document
    lock_button.setChecked(checked)
    on_toggle_lock(checked, noconvert=True)
    lock_button.triggered.connect(on_toggle_lock)
    set_properties_editor(statuswidget)


def set_properties_editor(statuswidget):
    """Adds additional buttons to the properties editor"""

    if hasattr(statuswidget, "propertybuttons"):
        statuswidget.propertybuttons.show()
    else:
        from PySide import QtCore, QtGui  # lazy loading

        mw = FreeCADGui.getMainWindow()
        editor = mw.findChild(QtGui.QTabWidget, "propertyTab")
        if editor:
            pTabCornerWidget = QtGui.QWidget()
            pButton1 = QtGui.QToolButton(pTabCornerWidget)
            pButton1.setText("")
            pButton1.setToolTip(translate("BIM", "Add IFC property..."))
            pButton1.setIcon(QtGui.QIcon(":/icons/IFC.svg"))
            pButton1.clicked.connect(on_add_property)
            pButton2 = QtGui.QToolButton(pTabCornerWidget)
            pButton2.setText("")
            pButton2.setToolTip(translate("BIM", "Add standard IFC Property Set..."))
            pButton2.setIcon(QtGui.QIcon(":/icons/BIM_IfcProperties.svg"))
            pButton2.clicked.connect(on_add_pset)
            pHLayout = QtGui.QHBoxLayout(pTabCornerWidget)
            pHLayout.addWidget(pButton1)
            pHLayout.addWidget(pButton2)
            pHLayout.setSpacing(2)
            pHLayout.setContentsMargins(2, 2, 0, 0)
            pHLayout.insertStretch(0)
            editor.setCornerWidget(pTabCornerWidget, QtCore.Qt.BottomRightCorner)
            statuswidget.propertybuttons = pTabCornerWidget
            QtCore.QTimer.singleShot(0, pTabCornerWidget.show)


def on_add_property():
    """When the 'add property' button is clicked"""

    sel = FreeCADGui.Selection.getSelection()
    if not sel:
        return
    from PySide import QtGui  # lazy loading
    from . import ifc_psets

    obj = sel[0]
    psets = list(set([obj.getGroupOfProperty(p) for p in obj.PropertiesList]))
    psets = [p for p in psets if p]
    psets = [p for p in psets if p not in ["Base", "IFC", "Geometry"]]
    mw = FreeCADGui.getMainWindow()
    editor = mw.findChild(QtGui.QTabWidget, "propertyTab")
    pset = None
    if editor:
        wid = editor.currentWidget()
        if wid and wid.objectName() == "propertyEditorData":
            if wid.currentIndex().parent():
                pset = wid.currentIndex().parent().data()
            else:
                pset = wid.currentIndex().data()
    form = FreeCADGui.PySideUic.loadUi(":/ui/dialogAddProperty.ui")
    # center the dialog over FreeCAD window
    form.move(mw.frameGeometry().topLeft() + mw.rect().center() - form.rect().center())
    form.field_pset.clear()
    form.field_pset.addItems(psets)
    if pset and (pset in psets):
        form.field_pset.setCurrentIndex(psets.index(pset))
    # TODO check for name duplicates while typing
    # execute
    result = form.exec_()
    if not result:
        return
    pname = form.field_name.text()
    if pname in obj.PropertiesList:
        print("DEBUG: property already exists", pname)
        return
    pset = form.field_pset.currentText()
    if not pset:
        # TODO disable the OK button if empty
        t = translate("BIM", "No Property set provided")
        FreeCAD.Console.PrintError(t + "\n")
    ptype = form.field_type.currentIndex()
    ptype = [
        "IfcLabel",
        "IfcBoolean",
        "IfcInteger",
        "IfcReal",
        "IfcLengthMeasure",
        "IfcAreaMeasure",
    ][ptype]
    fctype = ifc_psets.get_freecad_type(ptype)
    FreeCAD.ActiveDocument.openTransaction(translate("BIM", "add property"))
    for obj in sel:
        obj.addProperty(fctype, pname, pset, ptype + ":" + pname)
        ifc_psets.edit_pset(obj, pname, force=True)
    FreeCAD.ActiveDocument.commitTransaction()


def on_add_pset():
    """When the 'add pset' button is pressed"""

    def read_csv(csvfile):
        result = {}
        if os.path.exists(csvfile):
            with open(csvfile, "r") as f:
                reader = csv.reader(f, delimiter=";")
                for row in reader:
                    result[row[0]] = row[1:]
        return result

    def get_fcprop(ifcprop):
        if ifcprop == "IfcLengthMeasure":
            return "App::PropertyDistance"
        elif ifcprop == "IfcPositiveLengthMeasure":
            return "App::PropertyLength"
        elif ifcprop in ["IfcBoolean", "IfcLogical"]:
            return "App::PropertyBool"
        elif ifcprop == "IfcInteger":
            return "App::PropertyInteger"
        elif ifcprop == "IfcReal":
            return "App::PropertyFloat"
        elif ifcprop == "IfcAreaMeasure":
            return "App::PropertyArea"
        return "App::PropertyString"

    sel = FreeCADGui.Selection.getSelection()
    if not sel:
        return
    from . import ifc_psets

    obj = sel[0]
    mw = FreeCADGui.getMainWindow()
    # read standard psets
    psetpath = os.path.join(
        FreeCAD.getResourceDir(), "Mod", "BIM", "Presets", "pset_definitions.csv"
    )
    custompath = os.path.join(FreeCAD.getUserAppDataDir(), "BIM", "CustomPsets.csv")
    psetdefs = read_csv(psetpath)
    psetdefs.update(read_csv(custompath))
    psetkeys = list(psetdefs.keys())
    psetkeys.sort()
    form = FreeCADGui.PySideUic.loadUi(":/ui/dialogAddPSet.ui")
    # center the dialog over FreeCAD window
    form.move(mw.frameGeometry().topLeft() + mw.rect().center() - form.rect().center())
    form.field_pset.clear()
    form.field_pset.addItems(psetkeys)
    # execute
    result = form.exec_()
    if not result:
        return
    pset = form.field_pset.currentText()
    existing_psets = list(set([obj.getGroupOfProperty(p) for p in obj.PropertiesList]))
    if pset in existing_psets:
        t = translate("BIM", "Property set already exists")
        FreeCAD.Console.PrintError(t + ": " + pset + "\n")
        return
    props = [psetdefs[pset][i : i + 2] for i in range(0, len(psetdefs[pset]), 2)]
    props = [[p[0], p[1]] for p in props]
    FreeCAD.ActiveDocument.openTransaction(translate("BIM", "add property set"))
    for obj in sel:
        existing_psets = list(set([obj.getGroupOfProperty(p) for p in obj.PropertiesList]))
        if pset not in existing_psets:
            ifc_psets.add_pset(obj, pset)
        for prop in props:
            if prop[0] in obj.PropertiesList:
                t = translate("BIM", "Property already exists")
                FreeCAD.Console.PrintWarning(t + ": " + obj.Label + "," + prop[0] + "\n")
            else:
                obj.addProperty(get_fcprop(prop[1]), prop[0], pset, prop[1] + ":" + prop[0])
    FreeCAD.ActiveDocument.commitTransaction()


def on_toggle_lock(checked=None, noconvert=False, setchecked=False):
    """When the toolbar button is pressed"""

    if checked is None:
        checked = get_lock_status()
    set_menu(checked)
    set_button(checked, setchecked)
    if not noconvert:
        if checked:
            lock_document()
        else:
            unlock_document()


def on_open():
    """What happens when opening an existing document"""

    pass  # TODO implement


def on_activate():
    """What happens when activating a document"""

    from PySide import QtGui  # lazy import

    # always reset the menu to normal first
    set_menu(False)
    if FreeCADGui.activeWorkbench().name() != "BIMWorkbench":
        return
    doc = FreeCAD.ActiveDocument
    if doc and "IfcFilePath" in doc.PropertiesList:
        checked = True
    else:
        checked = False
    mw = FreeCADGui.getMainWindow()
    statuswidget = mw.findChild(QtGui.QToolBar, "BIMStatusWidget")
    if hasattr(statuswidget, "lock_button"):
        statuswidget.lock_button.setChecked(checked)
    on_toggle_lock(checked, noconvert=True)


def on_new():
    """What happens when creating a new document"""

    pass  # TODO implement


def set_menu(locked=False):
    """Sets the File menu items"""

    from PySide import QtGui  # lazy loading

    # switch Std_Save and IFC_Save
    mw = FreeCADGui.getMainWindow()
    wb = FreeCADGui.activeWorkbench()
    save_action = mw.findChild(QtGui.QAction, "Std_Save")
    if locked and "IFC_Save" in FreeCADGui.listCommands():
        if not hasattr(FreeCADGui, "IFC_WBManipulator"):
            FreeCADGui.IFC_WBManipulator = IFC_WBManipulator()
        # we need to void the shortcut otherwise it keeps active
        # even if the command is not shown
        FreeCADGui.IFC_saveshortcut = save_action.shortcut()
        save_action.setShortcut("")
        FreeCADGui.addWorkbenchManipulator(FreeCADGui.IFC_WBManipulator)
        wb.reloadActive()
    else:
        if hasattr(FreeCADGui, "IFC_saveshortcut"):
            save_action.setShortcut(FreeCADGui.IFC_saveshortcut)
            del FreeCADGui.IFC_saveshortcut
        if hasattr(FreeCADGui, "IFC_WBManipulator"):
            FreeCADGui.removeWorkbenchManipulator(FreeCADGui.IFC_WBManipulator)
            del FreeCADGui.IFC_WBManipulator
        wb.reloadActive()


def set_button(checked=False, setchecked=False):
    """Sets the lock button"""

    from PySide import QtGui  # lazy loading

    mw = FreeCADGui.getMainWindow()
    statuswidget = mw.findChild(QtGui.QToolBar, "BIMStatusWidget")
    if hasattr(statuswidget, "lock_button"):
        lock_button = statuswidget.lock_button
        if checked:
            lock_button.setToolTip(text_on)
            icon = QtGui.QIcon(":/icons/IFC.svg")
            lock_button.setIcon(icon)
            if setchecked:
                lock_button.setChecked(True)
        else:
            lock_button.setToolTip(text_off)
            image = QtGui.QImage(":/icons/IFC.svg")
            grayscale = image.convertToFormat(QtGui.QImage.Format_Grayscale8)
            grayscale = grayscale.convertToFormat(image.format())
            grayscale.setAlphaChannel(image)
            icon = QtGui.QIcon(QtGui.QPixmap.fromImage(grayscale))
            lock_button.setIcon(icon)
            if setchecked:
                lock_button.setChecked(False)


def unlock_document():
    """Unlocks the active document"""

    from . import ifc_tools  # lazy loading

    doc = FreeCAD.ActiveDocument
    if not doc:
        return
    if "IfcFilePath" in doc.PropertiesList:
        # this is a locked document
        doc.openTransaction("Unlock document")
        children = [o for o in doc.Objects if not o.InList]
        project = None
        if children:
            project = ifc_tools.create_document_object(doc, filename=doc.IfcFilePath, silent=True)
            project.Group = children
        props = ["IfcFilePath", "Modified", "Proxy", "Schema"]
        props += [p for p in doc.PropertiesList if doc.getGroupOfProperty(p) == "IFC"]
        for prop in props:
            doc.setPropertyStatus(prop, "-LockDynamic")
            doc.removeProperty(prop)
        if project:
            project.Modified = True
        doc.commitTransaction()
        doc.recompute()


def lock_document():
    """Locks the active document"""

    from . import ifc_tools  # lazy loading
    from importers import exportIFC
    from . import ifc_geometry
    from . import ifc_export
    from PySide import QtCore

    doc = FreeCAD.ActiveDocument
    if not doc:
        return
    products = []
    spatial = []
    ifcfile = None
    if "IfcFilePath" not in doc.PropertiesList:
        # this is not a locked document
        projects = [o for o in doc.Objects if getattr(o, "Class", None) == "IfcProject"]
        if len(projects) == 1:
            # 1 there is a project already
            project = projects[0]
            children = project.OutListRecursive
            rest = [o for o in doc.Objects if o not in children and o != project]
            doc.openTransaction("Lock document")
            ifc_tools.convert_document(doc, filename=project.IfcFilePath, strategy=3, silent=True)
            ifcfile = doc.Proxy.ifcfile
            if rest:
                # 1b some objects are outside
                objs = find_toplevel(rest)
                prefs, context = ifc_export.get_export_preferences(ifcfile)
                products = exportIFC.export(objs, ifcfile, preferences=prefs)
                for product in products.values():
                    if not getattr(product, "ContainedInStructure", None):
                        if not getattr(product, "FillsVoids", None):
                            if not getattr(product, "VoidsElements", None):
                                if not getattr(product, "Decomposes", None):
                                    new = ifc_tools.create_object(product, doc, ifcfile)
                                    children = ifc_tools.create_children(
                                        new, ifcfile, recursive=True
                                    )
                                    for o in [new] + children:
                                        ifc_geometry.add_geom_properties(o)
                for n in [o.Name for o in rest]:
                    doc.removeObject(n)
            else:
                # 1a all objects are already inside a project
                pass
            doc.removeObject(project.Name)
            doc.Modified = True
            # all objects have been deleted, we need to show at least something
            if not doc.Objects:
                ifc_tools.create_children(doc, ifcfile, recursive=True)
            doc.commitTransaction()
            doc.recompute()
        elif len(projects) > 1:
            # 2 there is more than one project
            FreeCAD.Console.PrintError(
                "Unable to lock this document because it contains several IFC documents\n"
            )
            QtCore.QTimer.singleShot(100, on_toggle_lock)
        elif doc.Objects:
            # 3 there is no project but objects
            doc.openTransaction("Lock document")
            objs = find_toplevel(doc.Objects)
            deletelist = [o.Name for o in doc.Objects]
            # ifc_export.export_and_convert(objs, doc)
            ifc_export.direct_conversion(objs, doc)
            for n in deletelist:
                if doc.getObject(n):
                    doc.removeObject(n)
            doc.IfcFilePath = ""
            doc.Modified = True
            doc.commitTransaction()
            doc.recompute()
        else:
            # 4 this is an empty document
            doc.openTransaction("Create IFC document")
            ifc_tools.convert_document(doc)
            doc.commitTransaction()
            doc.recompute()
        # reveal file contents if needed
        if "IfcFilePath" in doc.PropertiesList:
            create = True
            for o in doc.Objects:
                # scan for site or building
                if getattr(o, "IfcClass", "") in ("IfcSite", "IfcBuilding"):
                    create = False
                    break
            if create:
                if not ifcfile:
                    ifcfile = doc.Proxy.ifcfile
                ifc_tools.create_children(doc, ifcfile, recursive=False)


def find_toplevel(objs):
    """Finds the top-level objects from the list"""

    # filter out any object that depend on another from the list
    nobjs = []
    for obj in objs:
        for parent in obj.InListRecursive:
            if parent in objs:
                # exception: The object is hosting another
                if hasattr(parent, "Host") and parent.Host == obj:
                    nobjs.append(obj)
                elif hasattr(parent, "Hosts") and obj in parent.Hosts:
                    nobjs.append(obj)
                break
        else:
            nobjs.append(obj)
    # filter out non-convertible objects
    objs = filter_out(nobjs)
    return objs


def filter_out(objs):
    """Filter out objects that should not be converted to IFC"""

    import Draft

    nobjs = []
    for obj in objs:
        if obj.isDerivedFrom("Part::Feature"):
            nobjs.append(obj)
        elif obj.isDerivedFrom("Mesh::Feature"):
            nobjs.append(obj)
        elif Draft.is_group(obj):
            if filter_out(obj.Group):
                # only append groups that contain exportable objects
                nobjs.append(obj)
            else:
                print("DEBUG: Filtering out", obj.Label)
        elif obj.isDerivedFrom("App::Feature"):
            if Draft.get_type(obj) in (
                "Dimension",
                "LinearDimension",
                "Layer",
                "Text",
                "DraftText",
            ):
                nobjs.append(obj)
            else:
                print("DEBUG: Filtering out", obj.Label)
        else:
            print("DEBUG: Filtering out", obj.Label)
    return nobjs


def get_lock_status():
    """Returns the status of the IFC lock button"""

    if not FreeCAD.GuiUp:
        return PARAMS.GetBool("SingleDoc")
    from PySide import QtGui

    mw = FreeCADGui.getMainWindow()
    statuswidget = mw.findChild(QtGui.QToolBar, "BIMStatusWidget")
    if hasattr(statuswidget, "lock_button"):
        return statuswidget.lock_button.isChecked()
    else:
        return False


# add entry to File menu
# https://github.com/FreeCAD/FreeCAD/pull/10933
class IFC_WBManipulator:
    def modifyMenuBar(self):
        return [
            {"remove": "Std_Save"},
            {"remove": "Std_SaveAs"},
            {"insert": "IFC_Save", "menuItem": "Std_SaveCopy"},
            {"insert": "IFC_SaveAs", "menuItem": "Std_SaveCopy"},
        ]

    def modifyToolBars(self):
        return [
            {"remove": "Std_Save"},
            {"append": "IFC_Save", "toolBar": "File"},
        ]



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create PathTwistedArray objects.

The copies will be created along a path, like a polyline, B-spline,
or Bezier curve.
"""
## @package gui_pathtwistedarray
# \ingroup draftguitools
# \brief Provides GUI tools to create PathTwistedArray objects.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import Draft_rc
import draftguitools.gui_base_original as gui_base_original

from draftutils.messages import _err
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class PathTwistedArray(gui_base_original.Modifier):
    """Gui Command for the Path twisted array tool.

    Parameters
    ----------
    use_link: bool, optional
        It defaults to `False`. If it is `True`, the created object
        will be a `Link array`.
    """

    def __init__(self, use_link=False):
        super(PathTwistedArray, self).__init__()
        self.use_link = use_link
        self.call = None

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_PathTwistedArray",
            "MenuText": QT_TRANSLATE_NOOP("Draft_PathTwistedArray", "Twisted Path Array"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_PathTwistedArray",
                "Creates twisted copies of the selected object along a selected path",
            ),
        }

    def Activated(self, name="Path twisted array"):
        """Execute when the command is called."""
        super(PathTwistedArray, self).Activated(name=name)
        self.name = name
        self.proceed()

    def proceed(self):
        """Proceed with the command if one object was selected."""
        sel = Gui.Selection.getSelectionEx()
        if len(sel) != 2:
            _err(
                translate(
                    "draft",
                    "Select exactly 2 objects, the base object and the path object, before calling this command",
                )
            )
        else:
            base_object = sel[0].Object
            path_object = sel[1].Object

            count = 15
            rot_factor = 0.25
            use_link = self.use_link

            Gui.addModule("Draft")
            _cmd = "Draft.make_path_twisted_array"
            _cmd += "("
            _cmd += "App.ActiveDocument." + base_object.Name + ", "
            _cmd += "App.ActiveDocument." + path_object.Name + ", "
            _cmd += "count=" + str(count) + ", "
            _cmd += "rot_factor=" + str(rot_factor) + ", "
            _cmd += "use_link=" + str(use_link)
            _cmd += ")"

            _cmd_list = [
                "_obj_ = " + _cmd,
                "Draft.autogroup(_obj_)",
                "App.ActiveDocument.recompute()",
            ]
            self.commit(translate("draft", "Create Path Twisted Array"), _cmd_list)

        # Commit the transaction and execute the commands
        # through the parent class
        self.finish()


Gui.addCommand("Draft_PathTwistedArray", PathTwistedArray())


class PathTwistedLinkArray(PathTwistedArray):
    """Gui Command for the PathTwistedLinkArray tool."""

    def __init__(self):
        super(PathTwistedLinkArray, self).__init__(use_link=True)

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_PathTwistedLinkArray",
            "MenuText": QT_TRANSLATE_NOOP("Draft_PathTwistedLinkArray", "Twisted Path Link Array"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_PathTwistedLinkArray",
                "Creates twisted linked copies of the selected object along a selected path",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super(PathTwistedLinkArray, self).Activated(name="Path twisted link array")


Gui.addCommand("Draft_PathTwistedLinkArray", PathTwistedLinkArray())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2026 Mario Passaglia <mpassaglia[at]cbc.uba.ar>         *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import sys
from FreeCAD import Console
import Fem


def mesh_from_mesher(femmesh_obj, mesher=""):
    tool = None
    success = False
    match mesher:
        case "netgen":
            from femmesh import netgentools
            tool = netgentools.NetgenTools(femmesh_obj)
        case "gmsh" | "":
            from femmesh import gmshtools
            tool = gmshtools.GmshTools(femmesh_obj)
        case _:
            raise ValueError(f"Invalid mesher: {mesher}")

    # Redirect process error to report view
    print_error = lambda: Console.PrintError(
        tool.process.readAllStandardError().data().decode("utf-8")
    )
    tool.process.readyReadStandardError.connect(print_error)

    # generate the mesh
    try:
        success = tool.run(blocking=True)
    except Exception as e:
        error = sys.exc_info()[1]
        Console.PrintError(f"Unexpected error when creating mesh: {error}\n")

    return success

def mesh_from_existing(create_nodes, create_elements):
    fem_mesh = Fem.FemMesh()
    control = create_nodes(fem_mesh)
    if not control:
        Console.PrintError("Error on creating nodes.\n")
    control = create_elements(fem_mesh)
    if not control:
        Console.PrintError("Error on creating elements.\n")

    return fem_mesh



# **************************************************************************
#   Copyright (c) 2011 Juergen Riegel <FreeCAD@juergen-riegel.net>        *
#                                                                         *
#   This file is part of the FreeCAD CAx development system.              *
#                                                                         *
#   This program is free software; you can redistribute it and/or modify  *
#   it under the terms of the GNU Lesser General Public License (LGPL)    *
#   as published by the Free Software Foundation; either version 2 of     *
#   the License, or (at your option) any later version.                   *
#   for detail see the LICENCE text file.                                 *
#                                                                         *
#   FreeCAD is distributed in the hope that it will be useful,            *
#   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
#   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
#   GNU Library General Public License for more details.                  *
#                                                                         *
#   You should have received a copy of the GNU Library General Public     *
#   License along with FreeCAD; if not, write to the Free Software        *
#   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
#   USA                                                                   *
# **************************************************************************


import os, tempfile, unittest
import FreeCAD, Part, Sketcher
from Part import Precision

App = FreeCAD

xy_normal = FreeCAD.Vector(0, 0, 1)


def vec(x, y):
    """Shorthand to create a vector in the XY-plane"""
    return FreeCAD.Vector(x, y, 0)


def CreateRectangleSketch(SketchFeature, corner, lengths):
    hmin, hmax = corner[0], corner[0] + lengths[0]
    vmin, vmax = corner[1], corner[1] + lengths[1]

    # add the geometry and grab the count offset
    i = int(SketchFeature.GeometryCount)
    SketchFeature.addGeometry(
        Part.LineSegment(FreeCAD.Vector(hmin, vmax), FreeCAD.Vector(hmax, vmax, 0))
    )
    SketchFeature.addGeometry(
        Part.LineSegment(FreeCAD.Vector(hmax, vmax, 0), FreeCAD.Vector(hmax, vmin, 0))
    )
    SketchFeature.addGeometry(
        Part.LineSegment(FreeCAD.Vector(hmax, vmin, 0), FreeCAD.Vector(hmin, vmin, 0))
    )
    SketchFeature.addGeometry(
        Part.LineSegment(FreeCAD.Vector(hmin, vmin, 0), FreeCAD.Vector(hmin, vmax, 0))
    )

    # add the rectangular constraints
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", i + 0, 2, i + 1, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", i + 1, 2, i + 2, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", i + 2, 2, i + 3, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", i + 3, 2, i + 0, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Horizontal", i + 0))
    SketchFeature.addConstraint(Sketcher.Constraint("Horizontal", i + 2))
    SketchFeature.addConstraint(Sketcher.Constraint("Vertical", i + 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Vertical", i + 3))

    # Fix the bottom left corner of the rectangle
    SketchFeature.addConstraint(Sketcher.Constraint("DistanceX", i + 2, 2, corner[0]))
    SketchFeature.addConstraint(Sketcher.Constraint("DistanceY", i + 2, 2, corner[1]))

    # add dimensions
    if lengths[0] == lengths[1]:
        SketchFeature.addConstraint(Sketcher.Constraint("Equal", i + 2, i + 3))
        SketchFeature.addConstraint(Sketcher.Constraint("Distance", i + 0, hmax - hmin))
    else:
        SketchFeature.addConstraint(Sketcher.Constraint("Distance", i + 1, vmax - vmin))
        SketchFeature.addConstraint(Sketcher.Constraint("Distance", i + 0, hmax - hmin))


def CreateCircleSketch(SketchFeature, center, radius):
    i = int(SketchFeature.GeometryCount)
    SketchFeature.addGeometry(Part.Circle(App.Vector(*center), App.Vector(0, 0, 1), radius), False)
    SketchFeature.addConstraint(Sketcher.Constraint("Radius", i, radius))
    SketchFeature.addConstraint(Sketcher.Constraint("DistanceX", i, 3, center[0]))
    SketchFeature.addConstraint(Sketcher.Constraint("DistanceY", i, 3, center[1]))


def CreateBoxSketchSet(SketchFeature):
    SketchFeature.addGeometry(
        Part.LineSegment(
            FreeCAD.Vector(-99.230339, 36.960674, 0),
            FreeCAD.Vector(69.432587, 36.960674, 0),
        )
    )
    SketchFeature.addGeometry(
        Part.LineSegment(
            FreeCAD.Vector(69.432587, 36.960674, 0),
            FreeCAD.Vector(69.432587, -53.196629, 0),
        )
    )
    SketchFeature.addGeometry(
        Part.LineSegment(
            FreeCAD.Vector(69.432587, -53.196629, 0),
            FreeCAD.Vector(-99.230339, -53.196629, 0),
        )
    )
    SketchFeature.addGeometry(
        Part.LineSegment(
            FreeCAD.Vector(-99.230339, -53.196629, 0),
            FreeCAD.Vector(-99.230339, 36.960674, 0),
        )
    )
    # add the constraints
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Horizontal", 0))
    SketchFeature.addConstraint(Sketcher.Constraint("Horizontal", 2))
    SketchFeature.addConstraint(Sketcher.Constraint("Vertical", 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Vertical", 3))
    # add dimensions
    SketchFeature.addConstraint(Sketcher.Constraint("Distance", 1, 81.370787))
    SketchFeature.addConstraint(Sketcher.Constraint("Distance", 0, 187.573036))


def CreateSlotPlateSet(SketchFeature):
    SketchFeature.addGeometry(
        Part.LineSegment(
            App.Vector(60.029362, -30.279360, 0), App.Vector(-120.376335, -30.279360, 0)
        )
    )
    SketchFeature.addConstraint(Sketcher.Constraint("Horizontal", 0))
    SketchFeature.addGeometry(
        Part.LineSegment(
            App.Vector(-120.376335, -30.279360, 0), App.Vector(-70.193062, 38.113884, 0)
        )
    )
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
    SketchFeature.addGeometry(
        Part.LineSegment(App.Vector(-70.193062, 38.113884, 0), App.Vector(60.241116, 37.478645, 0))
    )
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Horizontal", 2))
    SketchFeature.addGeometry(
        Part.ArcOfCircle(
            Part.Circle(App.Vector(60.039921, 3.811391, 0), App.Vector(0, 0, 1), 35.127132),
            -1.403763,
            1.419522,
        )
    )
    SketchFeature.addConstraint(Sketcher.Constraint("Tangent", 2, 2, 3, 2))
    SketchFeature.addConstraint(Sketcher.Constraint("Tangent", 0, 1, 3, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Angle", 0, 2, 1, 1, 0.947837))
    SketchFeature.addConstraint(Sketcher.Constraint("Distance", 0, 184.127425))
    SketchFeature.setDatum(7, 200.000000)
    SketchFeature.addConstraint(Sketcher.Constraint("Radius", 3, 38.424808))
    SketchFeature.setDatum(8, 40.000000)
    SketchFeature.setDatum(6, 0.872665)
    SketchFeature.addConstraint(Sketcher.Constraint("DistanceX", 0, 2, 0.0))
    SketchFeature.setDatum(9, 0.000000)
    SketchFeature.moveGeometry(0, 2, App.Vector(-0.007829, -33.376450, 0))
    SketchFeature.moveGeometry(0, 2, App.Vector(-0.738149, -10.493386, 0))
    SketchFeature.moveGeometry(0, 2, App.Vector(-0.007829, 2.165328, 0))
    SketchFeature.addConstraint(Sketcher.Constraint("DistanceY", 0, 2, 2.165328))
    SketchFeature.setDatum(10, 0.000000)


def CreateSlotPlateInnerSet(SketchFeature):
    SketchFeature.addGeometry(
        Part.Circle(App.Vector(195.055893, 39.562252, 0), App.Vector(0, 0, 1), 29.846098)
    )
    SketchFeature.addGeometry(
        Part.LineSegment(App.Vector(150.319031, 13.449363, 0), App.Vector(36.700474, 13.139774, 0))
    )
    SketchFeature.addConstraint(Sketcher.Constraint("Horizontal", 5))
    SketchFeature.addGeometry(
        Part.LineSegment(App.Vector(36.700474, 13.139774, 0), App.Vector(77.566010, 63.292927, 0))
    )
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 5, 2, 6, 1))
    SketchFeature.addGeometry(
        Part.LineSegment(App.Vector(77.566010, 63.292927, 0), App.Vector(148.151917, 63.602505, 0))
    )
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 6, 2, 7, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Horizontal", 7))
    SketchFeature.addConstraint(Sketcher.Constraint("Parallel", 1, 6))
    SketchFeature.addGeometry(
        Part.ArcOfCircle(
            Part.Circle(App.Vector(192.422913, 38.216347, 0), App.Vector(0, 0, 1), 45.315174),
            2.635158,
            3.602228,
        )
    )
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 7, 2, 8, 1))
    SketchFeature.addConstraint(Sketcher.Constraint("Coincident", 8, 2, 5, 1))


def CreateThreeLinesWithCommonPoint(SketchFeature):
    SketchFeature.addGeometry(
        Part.LineSegment(
            App.Vector(-55.965607, -9.864289, 0), App.Vector(-55.600571, -9.387639, 0)
        ),
        False,
    )
    SketchFeature.addGeometry(
        Part.LineSegment(
            App.Vector(-55.735817, -9.067246, 0), App.Vector(-55.600571, -9.387639, 0)
        ),
        False,
    )
    SketchFeature.addGeometry(
        Part.LineSegment(
            App.Vector(-55.600571, -9.387639, 0), App.Vector(-55.058266, -9.677831, 0)
        ),
        False,
    )


# ---------------------------------------------------------------------------
# define the test cases to test the FreeCAD Sketcher module
# ---------------------------------------------------------------------------


class TestSketcherSolver(unittest.TestCase):
    def setUp(self):
        self.Doc = FreeCAD.newDocument("SketchSolverTest")

    def testBoxCase(self):
        self.Box = self.Doc.addObject("Sketcher::SketchObject", "SketchBox")
        CreateBoxSketchSet(self.Box)
        self.Doc.recompute()
        # moving a point of the sketch
        self.Box.moveGeometry(0, 2, App.Vector(88.342697, 28.174158, 0))
        # fully constrain
        self.Box.addConstraint(Sketcher.Constraint("DistanceX", 1, 2, 90.0))
        self.Box.addConstraint(Sketcher.Constraint("DistanceY", 1, 2, -50.0))
        self.Doc.recompute()

    def testSlotCase(self):
        self.Slot = self.Doc.addObject("Sketcher::SketchObject", "SketchSlot")
        CreateSlotPlateSet(self.Slot)
        self.Doc.recompute()
        # test if all edges created
        self.assertTrue(len(self.Slot.Shape.Edges) == 4)
        CreateSlotPlateInnerSet(self.Slot)
        self.Doc.recompute()
        self.assertTrue(len(self.Slot.Shape.Edges) == 9)

    def testIssue3245(self):
        self.Doc2 = FreeCAD.newDocument("Issue3245")
        self.Doc2.addObject("Sketcher::SketchObject", "Sketch")
        self.Doc2.Sketch.Placement = App.Placement(
            App.Vector(0.000000, 0.000000, 0.000000),
            App.Rotation(0.000000, 0.000000, 0.000000, 1.000000),
        )
        self.Doc2.Sketch.MapMode = "Deactivated"
        self.Doc2.Sketch.addGeometry(
            Part.LineSegment(
                App.Vector(-1.195999, 56.041161, 0), App.Vector(60.654316, 56.382877, 0)
            ),
            False,
        )
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("PointOnObject", 0, 1, -2))
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("Horizontal", 0))
        self.Doc2.Sketch.addGeometry(
            Part.LineSegment(
                App.Vector(0.512583, 32.121155, 0), App.Vector(60.654316, 31.779440, 0)
            ),
            False,
        )
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("Horizontal", 1))
        self.Doc2.Sketch.addGeometry(
            Part.LineSegment(
                App.Vector(0.170867, 13.326859, 0), App.Vector(61.679455, 13.326859, 0)
            ),
            False,
        )
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("PointOnObject", 2, 1, -2))
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("Horizontal", 2))
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("PointOnObject", 1, 1, -2))
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("DistanceX", 0, 1, 0, 2, 60.654316))
        self.Doc2.Sketch.setExpression("Constraints[6]", "60")
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("DistanceX", 1, 1, 1, 2, 60.654316))
        self.Doc2.Sketch.setExpression("Constraints[7]", "65")
        self.Doc2.Sketch.addConstraint(Sketcher.Constraint("DistanceX", 2, 1, 2, 2, 61.679455))
        self.Doc2.Sketch.setExpression("Constraints[8]", "70")
        self.Doc2.recompute()
        self.Doc2.Sketch.delGeometry(2)
        values = d = {key: value for (key, value) in self.Doc2.Sketch.ExpressionEngine}
        self.assertTrue(values["Constraints[4]"] == "60")
        self.assertTrue(values["Constraints[5]"] == "65")
        FreeCAD.closeDocument("Issue3245")

    def testIssue3245_2(self):
        self.Doc2 = FreeCAD.newDocument("Issue3245")
        ActiveSketch = self.Doc2.addObject("Sketcher::SketchObject", "Sketch")
        ActiveSketch.Placement = App.Placement(
            App.Vector(0.000000, 0.000000, 0.000000),
            App.Rotation(0.000000, 0.000000, 0.000000, 1.000000),
        )
        ActiveSketch.MapMode = "Deactivated"
        geoList = []
        geoList.append(
            Part.LineSegment(
                App.Vector(-23.574591, 42.399727, 0),
                App.Vector(81.949776, 42.399727, 0),
            )
        )
        geoList.append(
            Part.LineSegment(
                App.Vector(81.949776, 42.399727, 0),
                App.Vector(81.949776, -19.256901, 0),
            )
        )
        geoList.append(
            Part.LineSegment(
                App.Vector(81.949776, -19.256901, 0),
                App.Vector(-23.574591, -19.256901, 0),
            )
        )
        geoList.append(
            Part.LineSegment(
                App.Vector(-23.574591, -19.256901, 0),
                App.Vector(-23.574591, 42.399727, 0),
            )
        )
        ActiveSketch.addGeometry(geoList, False)
        conList = []
        conList.append(Sketcher.Constraint("Coincident", 0, 2, 1, 1))
        conList.append(Sketcher.Constraint("Coincident", 1, 2, 2, 1))
        conList.append(Sketcher.Constraint("Coincident", 2, 2, 3, 1))
        conList.append(Sketcher.Constraint("Coincident", 3, 2, 0, 1))
        conList.append(Sketcher.Constraint("Horizontal", 0))
        conList.append(Sketcher.Constraint("Horizontal", 2))
        conList.append(Sketcher.Constraint("Vertical", 1))
        conList.append(Sketcher.Constraint("Vertical", 3))
        ActiveSketch.addConstraint(conList)
        ActiveSketch.addConstraint(Sketcher.Constraint("DistanceX", 0, 1, 0, 2, 105.524367))
        ActiveSketch.setExpression("Constraints[8]", "10 + 10")
        ActiveSketch.addConstraint(Sketcher.Constraint("DistanceY", 3, 1, 3, 2, 61.656628))
        ActiveSketch.setDatum(9, App.Units.Quantity("5.000000 mm"))
        ActiveSketch.delConstraint(8)
        values = d = {key: value for (key, value) in self.Doc2.Sketch.ExpressionEngine}
        self.Doc2.recompute()
        self.assertTrue(len(values) == 0)
        FreeCAD.closeDocument("Issue3245")

    def testBlockConstraintEllipse(self):
        self.Doc3 = FreeCAD.newDocument("BlockConstraintTests")
        ActiveSketch = self.Doc3.addObject("Sketcher::SketchObject", "Sketch")
        ActiveSketch.Placement = App.Placement(
            App.Vector(0.000000, 0.000000, 0.000000),
            App.Rotation(0.000000, 0.000000, 0.000000, 1.000000),
        )
        ActiveSketch.MapMode = "Deactivated"
        ActiveSketch.addGeometry(
            Part.Ellipse(
                App.Vector(-19.129438, 14.345055, 0),
                App.Vector(-33.806261, 12.085921, 0),
                App.Vector(-30.689360, 7.107538, 0),
            ),
            False,
        )
        ActiveSketch.solve()
        ActiveSketch.exposeInternalGeometry(0)
        ActiveSketch.solve()
        ActiveSketch.moveGeometry(0, 0, App.Vector(-26.266434, 14.345055, 0), 0)
        ActiveSketch.solve()
        ActiveSketch.addConstraint(Sketcher.Constraint("Block", 0))  # Block the Ellipse in place
        ActiveSketch.addConstraint(
            Sketcher.Constraint("Block", 1)
        )  # Block the major axis in place (on purpose)
        status = ActiveSketch.solve()
        self.assertTrue(status == 0)  # no redundants/conflicts/convergence issues
        ActiveSketch.addConstraint(
            Sketcher.Constraint("Distance", 1, 27.277350)
        )  # Length of major axis
        ActiveSketch.setDriving(
            6, True
        )  # ensure length is driving (because pre-existing block constraint on major axis)
        ActiveSketch.setDatum(6, App.Units.Quantity("28.000000 mm"))
        status = ActiveSketch.solve()
        self.assertTrue(status == 0)  # no redundants/conflicts/convergence issues
        ActiveSketch.addConstraint(
            Sketcher.Constraint("Distance", 2, 11.747233)
        )  # Length of minor axis
        ActiveSketch.setDatum(7, App.Units.Quantity("15.000000 mm"))
        ActiveSketch.solve()
        self.assertTrue(status == 0)  # no redundants/conflicts/convergence issues
        ActiveSketch.addConstraint(Sketcher.Constraint("Block", 2))
        ActiveSketch.solve()
        self.assertTrue(status == 0)  # no redundants/conflicts/convergence issues
        ActiveSketch.addConstraint(
            Sketcher.Constraint("Horizontal", 1)
        )  # Make major axis horizontal (together with horizontal and length driving constraints)
        ActiveSketch.solve()
        self.assertTrue(status == 0)  # no redundants/conflicts/convergence issues
        ActiveSketch.addConstraint(
            Sketcher.Constraint("DistanceX", 0, 3, -1, 1, 27.655024)
        )  # Locate Ellipse center
        ActiveSketch.addConstraint(Sketcher.Constraint("DistanceY", 0, 3, -1, 1, -20.877021))
        ActiveSketch.setDatum(10, App.Units.Quantity("25.000000 mm"))
        ActiveSketch.setDatum(11, App.Units.Quantity("-20.000000 mm"))
        ActiveSketch.solve()
        self.assertTrue(status == 0)  # no redundants/conflicts/convergence issues
        FreeCAD.closeDocument(self.Doc3.Name)

    def testThreeLinesWithCoincidences_1(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        CreateThreeLinesWithCommonPoint(sketch)
        sketch.addConstraint(Sketcher.Constraint("Coincident", 1, 2, 0, 2))
        sketch.addConstraint(Sketcher.Constraint("Coincident", 2, 1, 0, 2))
        self.assertEqual(sketch.detectMissingPointOnPointConstraints(0.0001), 0)

    # Same as in testThreeLinesWithCoincidences_1 but set the constraints on
    # different lines
    def testThreeLinesWithCoincidences_2(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        CreateThreeLinesWithCommonPoint(sketch)
        sketch.addConstraint(Sketcher.Constraint("Coincident", 1, 2, 0, 2))
        sketch.addConstraint(Sketcher.Constraint("Coincident", 2, 1, 1, 2))
        self.assertEqual(sketch.detectMissingPointOnPointConstraints(0.0001), 0)

    def testCircleToLineDistance_Driving_Passant(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        radius = 20
        circle = Part.Circle(vec(0, 0), xy_normal, radius)
        line = Part.LineSegment(vec(-radius, 2 * radius), vec(radius, 2 * radius))
        c_idx = sketch.addGeometry(circle)
        l_idx = sketch.addGeometry(line)
        # use a positive distance, other than the initial distance of the line
        wanted_distance = radius / 2
        sketch.addConstraint(Sketcher.Constraint("Distance", c_idx, l_idx, wanted_distance))
        self.assertSuccessfulSolve(sketch)
        c_shape = sketch.Geometry[c_idx].toShape()
        l_shape = sketch.Geometry[l_idx].toShape()
        self.assertShapeDistance(c_shape, l_shape, wanted_distance)

    @unittest.skip("Support for secants still under discussion, see comments in PR 9044")
    def testCircleToLineDistance_Driving_Secant(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        radius = 20
        c_idx = sketch.addGeometry(Part.Circle(vec(0, 0), xy_normal, radius))
        l_idx = sketch.addGeometry(
            Part.LineSegment(vec(-radius, 2 * radius), vec(radius, 2 * radius))
        )
        # use a negative distance to tell "line is within the circle"
        wanted_distance = -radius / 2
        sketch.addConstraint(Sketcher.Constraint("Distance", c_idx, l_idx, wanted_distance))
        self.assertSuccessfulSolve(sketch)
        c_shape = sketch.Geometry[c_idx].toShape()
        l_shape = sketch.Geometry[l_idx].toShape()
        self.assertShapeDistance(c_shape, l_shape, 0)  # secant intersects circle, thus no distance

    def testCircleToLineDistance_Reference_Secant(self):
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        radius = 20
        c_idx = sketch.addGeometry(Part.Circle(vec(0, 0), xy_normal, radius))
        l_idx = sketch.addGeometry(
            Part.LineSegment(vec(-radius, radius / 2), vec(radius, radius / 2))
        )
        # The block constraints are required to ensure the geometry does not move.
        # Without this, the solver may find another valid solution than what we assert.
        sketch.addConstraint(
            [Sketcher.Constraint("Block", c_idx), Sketcher.Constraint("Block", l_idx)]
        )
        expected_distance = radius / 2  # note that we don't set this in the constraint below!
        # TODO: addConstraint(constraint) triggers a solve (for good reasons) however, this way
        # one cannot add non-driving constraints. In contrast, addConstraint(list(constraint))
        # does not solve automatically, thus we use this "overload".
        # Much nicer would be an addConstraint(constraint, isReference=False), like addGeometry
        dist_idx = sketch.addConstraint([Sketcher.Constraint("Distance", c_idx, l_idx, 0)])[0]
        sketch.setDriving(dist_idx, False)
        self.assertSuccessfulSolve(sketch)
        actual_distance = sketch.Constraints[dist_idx].Value
        self.assertAlmostEqual(
            expected_distance,
            actual_distance,
            delta=Precision.confusion(),
            msg="Reference constraint did not return the expected distance.",
        )

    def testCircleToLineDistance_Legacy_Negative(self):
        # compare a driving negative distance to an expected positive reference one
        sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
        radius = 20
        c_idx = sketch.addGeometry(Part.Circle(vec(0, 0), xy_normal, radius))
        l_idx = sketch.addGeometry(
            Part.LineSegment(vec(-radius, radius / 2), vec(radius, radius / 2))
        )
        sketch.addConstraint(
            [Sketcher.Constraint("Block", c_idx), Sketcher.Constraint("Block", l_idx)]
        )
        expected_distance = radius / 2
        ref_idx = sketch.addConstraint(
            [
                Sketcher.Constraint("Distance", c_idx, l_idx, -expected_distance),
                Sketcher.Constraint("Distance", c_idx, l_idx, 0),
            ]
        )[1]
        sketch.setDriving(ref_idx, False)
        self.assertSuccessfulSolve(sketch)
        actual_distance = sketch.Constraints[ref_idx].Value
        self.assertAlmostEqual(
            expected_distance,
            actual_distance,
            delta=Precision.confusion(),
            msg="Negative length constraint did not return the expected distance.",
        )

    def testRemovedExternalGeometryReference(self):
        if "BUILD_PARTDESIGN" in FreeCAD.__cmake__:
            body = self.Doc.addObject("PartDesign::Body", "Body")
            sketch = body.newObject("Sketcher::SketchObject", "Sketch")
            CreateRectangleSketch(sketch, (0, 0), (30, 30))
            pad = body.newObject("PartDesign::Pad", "Pad")
            pad.Profile = sketch
            sketch1 = body.newObject("Sketcher::SketchObject", "Sketch1")
            CreateCircleSketch(sketch1, (15, 15), 0.25)
            self.Doc.recompute()
            self.assertEqual(len(pad.Shape.Edges), 12)

            hole = self.Doc.addObject("PartDesign::Hole", "Hole")
            hole.Refine = True
            hole.Reversed = True
            body.addObject(hole)
            hole.Profile = sketch1
            hole.DrillPointAngle = 118.000000
            hole.Diameter = 6.000000
            hole.TaperedAngle = 90.000000
            hole.Tapered = 0
            hole.Depth = 8.000000
            hole.Threaded = 1
            hole.ModelThread = 0
            hole.ThreadDepthType = 0
            hole.ThreadType = 1
            hole.ThreadSize = 16
            hole.ThreadClass = 0
            hole.ThreadDirection = 0
            hole.HoleCutType = 0
            hole.DepthType = 0
            hole.DrillPoint = 1
            hole.DrillForDepth = 0
            self.Doc.recompute()
            # 15 edges if it's passthrough-flat 17 if DrillPoint = 1
            self.assertEqual(len(hole.Shape.Edges), 17)

            hole.ModelThread = 1
            sketch2 = body.newObject("Sketcher::SketchObject", "Sketch2")
            CreateRectangleSketch(sketch2, (0, 0), (3, 3))
            self.Doc.recompute()
            self.assertGreater(len(hole.Shape.Edges), 17)
            # 77 edges for basic profile
            self.assertEqual(len(hole.Shape.Edges), 77)

            # Edges in the thread should disappear when we stop modeling thread
            sketch2.addExternal("Hole", "Edge29")
            hole.ModelThread = 0
            hole.Refine = 1
            self.Doc.recompute()
            self.assertEqual(len(hole.Shape.Edges), 17)
            self.assertEqual(len(sketch2.ExternalGeometry), 0)

    def testSaveLoadWithExternalGeometryReference(self):
        if "BUILD_PARTDESIGN" in FreeCAD.__cmake__:
            # Arrange
            body = self.Doc.addObject("PartDesign::Body", "Body")
            sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
            CreateRectangleSketch(sketch, (0, 0), (30, 30))
            pad = self.Doc.addObject("PartDesign::Pad", "Pad")
            pad.Profile = sketch
            sketch1 = self.Doc.addObject("Sketcher::SketchObject", "Sketch1")
            body.addObject(sketch)
            body.addObject(pad)
            body.addObject(sketch1)
            self.Doc.recompute()
            sketch1.addExternal("Pad", "Edge12")
            self.Doc.recompute()

            # Act: Try changing sketch before the save
            sketch = self.Doc.getObject("Sketch")
            g1 = sketch.Constraints[11].First
            d1 = sketch.Constraints[11].Value
            sketch.delConstraint(11)
            sketch.addConstraint(Sketcher.Constraint("Distance", g1, d1 - 1.0))
            self.Doc.recompute()

            # Act: Save and reload the file
            filename = tempfile.gettempdir() + os.sep + self.Doc.Name + ".FCStd"
            self.Doc.saveAs(filename)
            FreeCAD.closeDocument(self.Doc.Name)
            self.Doc = FreeCAD.openDocument(filename)
            pad = self.Doc.getObject("Pad")
            sketch1 = self.Doc.getObject("Sketch1")
            self.Doc.recompute()

            # Act:  change sketch after restore ( trigger missing references if there is a bug )
            sketch = self.Doc.getObject("Sketch")
            g1 = sketch.Constraints[11].First
            d1 = sketch.Constraints[11].Value
            sketch.delConstraint(11)
            sketch.addConstraint(Sketcher.Constraint("Distance", g1, d1 - 1.0))
            self.Doc.recompute()

            # Assert
            self.assertEqual(pad.Shape.ElementMapSize, 30)
            self.assertIn("Edge12", pad.Shape.ElementReverseMap)
            self.assertIn((pad, ("Edge12",)), sketch1.ExternalGeometry)  # Not "?Edge12"

    def testTNPExternalGeometryStored(self):
        # Arrange
        if "BUILD_PARTDESIGN" in FreeCAD.__cmake__:
            import xml.etree.ElementTree as ET

            sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
            CreateRectangleSketch(sketch, (0, 0), (30, 30))
            sketch1 = self.Doc.addObject("Sketcher::SketchObject", "Sketch1")
            pad = self.Doc.addObject("PartDesign::Pad", "Pad")
            pad.Profile = sketch
            self.Doc.recompute()
            sketch1.addExternal("Pad", "Edge12")
            self.Doc.recompute()
            # Act
            root = ET.fromstring("<all>" + sketch1.Content + "</all>")
            # Can't use != in an xpath because it wasn't added until python 3.10.
            # "*/*[@name='ExternalGeo']//*/[@type='Sketcher::ExternalGeometryExtension']/[@Ref!='']"
            extRefs = root.findall(
                "*/*[@name='ExternalGeo']//*/[@type='Sketcher::ExternalGeometryExtension']/[@Ref='']"
            )
            extRefsAll = root.findall(
                "*/*[@name='ExternalGeo']//*/[@type='Sketcher::ExternalGeometryExtension']/[@Ref]"
            )
            # Assert
            self.assertEqual(len(extRefs), 2)
            self.assertEqual(len(extRefsAll), 3)
            self.assertEqual(root.tag, "all")
            # Act
            filename = tempfile.gettempdir() + os.sep + self.Doc.Name + ".FCStd"
            self.Doc.saveAs(filename)
            FreeCAD.closeDocument(self.Doc.Name)
            self.Doc = FreeCAD.openDocument(filename)
            # Assert
            root = ET.fromstring("<all>" + self.Doc.getObject("Sketch1").Content + "</all>")
            extRefs = root.findall(
                "*/*[@name='ExternalGeo']//*/[@type='Sketcher::ExternalGeometryExtension']/[@Ref='']"
            )
            extRefsAll = root.findall(
                "*/*[@name='ExternalGeo']//*/[@type='Sketcher::ExternalGeometryExtension']/[@Ref]"
            )
            self.assertEqual(len(extRefs), 2)
            self.assertEqual(len(extRefsAll), 3)
            self.assertEqual(root.tag, "all")
            # Act to change the constraint
            sketch = self.Doc.getObject("Sketch")
            g1 = sketch.Constraints[11].First
            d1 = sketch.Constraints[11].Value
            sketch.delConstraint(11)
            sketch.addConstraint(Sketcher.Constraint("Distance", g1, d1 - 1.0))
            self.Doc.recompute()
            # Assert
            root = ET.fromstring("<all>" + self.Doc.getObject("Sketch1").Content + "</all>")
            extRefs = root.findall(
                "*/*[@name='ExternalGeo']//*/[@type='Sketcher::ExternalGeometryExtension']/[@Ref='']"
            )
            extRefsAll = root.findall(
                "*/*[@name='ExternalGeo']//*/[@type='Sketcher::ExternalGeometryExtension']/[@Ref]"
            )
            self.assertEqual(len(extRefs), 2)
            self.assertEqual(len(extRefsAll), 3)
            self.assertEqual(root.tag, "all")

    def testConstructionToggleTNP(self):
        """Bug 15484"""
        if "BUILD_PARTDESIGN" in FreeCAD.__cmake__:
            # Arrange
            import xml.etree.ElementTree as ET

            body = self.Doc.addObject("PartDesign::Body", "Body")
            sketch = self.Doc.addObject("Sketcher::SketchObject", "Sketch")
            CreateRectangleSketch(sketch, (0, 0), (30, 30))
            # Add zigsag geo as construction lines
            i = sketch.GeometryCount
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(-10, 10, 0)), True
            )
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(-10, 10, 0), FreeCAD.Vector(-5, 20, 0)), True
            )
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(-5, 20, 0), FreeCAD.Vector(-10, 25, 0)), True
            )
            sketch.addGeometry(
                Part.LineSegment(FreeCAD.Vector(-10, 25, 0), FreeCAD.Vector(0, 30, 0)), True
            )
            sketch.addConstraint(Sketcher.Constraint("Coincident", i + 0, 2, i + 1, 1))
            sketch.addConstraint(Sketcher.Constraint("Coincident", i + 1, 2, i + 2, 1))
            sketch.addConstraint(Sketcher.Constraint("Coincident", i + 2, 2, i + 3, 1))
            sketch.addConstraint(Sketcher.Constraint("Coincident", i + 3, 2, 0, 1))
            sketch.addConstraint(Sketcher.Constraint("Coincident", i + 0, 1, 2, 2))

            pad = self.Doc.addObject("PartDesign::Pad", "Pad")
            pad.Profile = sketch
            body.addObject(sketch)
            body.addObject(pad)
            self.Doc.recompute()
            sketch1 = self.Doc.addObject("Sketcher::SketchObject", "Sketch1")
            sketch1.AttachmentSupport = (pad, ("Face6"))
            sketch1.MapMode = "FlatFace"
            self.Doc.recompute()

            CreateCircleSketch(sketch1, (2, 2, 0), 1)
            CreateCircleSketch(sketch1, (6, 2, 0), 1)
            body.addObject(sketch1)
            self.Doc.recompute()
            # Act toggle construction lines on in sketch; pad now has 9 instead of 6 faces.
            sketch.setConstruction(4, False)
            sketch.setConstruction(5, False)
            sketch.setConstruction(6, False)
            sketch.setConstruction(7, False)
            sketch.setConstruction(3, True)
            self.Doc.recompute()
            # Assert
            # AttachmentSupport is a list of (object,(subobject list)) with 1 entry.  Get the
            # first and only subobject name in second part of that first tuple and see that it moved
            # from the Face6 we set above.
            self.assertEqual(sketch1.AttachmentSupport[0][1][0], "Face9")
            self.assertIn("Face6", pad.Shape.ElementReverseMap)  # different Face6 exists

    # TODO other tests:
    #  getHigherElement

    def assertSuccessfulSolve(self, sketch, msg=None):
        status = sketch.solve()
        # TODO: can we get the solver's messages somehow to improve the message?
        self.assertTrue(status == 0, msg=msg or "solver didn't converge")

    def assertShapeDistance(self, shape1, shape2, expected_distance, msg=None):
        distance, _, _ = shape1.distToShape(shape2)
        self.assertAlmostEqual(
            distance,
            expected_distance,
            delta=Precision.confusion(),
            msg=msg or "The given shapes are not spaced by the expected distance.",
        )

    def tearDown(self):
        # closing doc
        FreeCAD.closeDocument("SketchSolverTest")
        # print ("omit closing document for debugging")



from Base.Metadata import export
from typing import Final
from Part.Geom2d import Curve2d

@export(
    PythonName="Part.Geom2d.Conic2d",
    Twin="Geom2dConic",
    TwinPointer="Geom2dConic",
    Include="Mod/Part/App/Geometry2d.h",
    FatherInclude="Mod/Part/App/Geom2d/Curve2dPy.h",
    Constructor=True,
)
class Conic2d(Curve2d):
    """
    Describes an abstract conic in 2d space

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    Location: object = ...
    """Location of the conic."""

    Eccentricity: Final[float] = ...
    """
    returns the eccentricity value of the conic e.
        e = 0 for a circle
        0 < e < 1 for an ellipse  (e = 0 if MajorRadius = MinorRadius)
        e > 1 for a hyperbola
        e = 1 for a parabola
    """

    XAxis: object = ...
    """The X axis direction of the circle"""

    YAxis: object = ...
    """The Y axis direction of the circle"""



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to create simple Text objects.

The textual block can consist of multiple lines.
"""
## @package gui_texts
# \ingroup draftguitools
# \brief Provides GUI tools to create simple Text objects.

## \addtogroup draftguitools
# @{
import sys
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
import DraftVecUtils
import draftguitools.gui_base_original as gui_base_original
import draftguitools.gui_tool_utils as gui_tool_utils

from draftutils.translate import translate
from draftutils.messages import _toolmsg

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Text(gui_base_original.Creator):
    """Gui command for the Text tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Text",
            "Accel": "T, E",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Text", "Text"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_Text", "Creates a multi-line annotation"),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Text")
        if self.ui:
            self.text = ""
            self.ui.sourceCmd = self
            self.ui.pointUi(title=translate("draft", self.featureName), icon="Draft_Text")
            self.ui.isRelative.hide()
            self.ui.continueCmd.show()
            self.call = self.view.addEventCallback("SoEvent", self.action)
            self.active = True
            self.ui.xValue.setFocus()
            self.ui.xValue.selectAll()
            _toolmsg(translate("draft", "Pick location point"))

    def finish(self, cont=False):
        """Terminate the operation.

        Parameters
        ----------
        cont: bool or None, optional
            Restart (continue) the command if `True`, or if `None` and
            `ui.continueMode` is `True`.
        """
        self.end_callbacks(self.call)
        super().finish(self)
        if cont or (cont is None and self.ui and self.ui.continueMode):
            self.Activated()

    def createObject(self):
        """Create the actual object in the current document."""
        rot, sup, pts, fil = self.getStrings()
        base = pts[1:-1]

        text_list = self.text

        if not text_list:
            self.finish()
            return None

        # If the last element is an empty string "" we remove it
        if not text_list[-1]:
            text_list.pop()

        t_list = ['"' + line + '"' for line in text_list]

        list_as_text = ", ".join(t_list)

        string = "[" + list_as_text + "]"

        Gui.addModule("Draft")
        _cmd = "Draft.make_text"
        _cmd += "("
        _cmd += string + ", "
        _cmd += "placement=pl, "
        _cmd += "screen=None, height=None, line_spacing=None"
        _cmd += ")"
        _cmd_list = [
            "pl = FreeCAD.Placement()",
            "pl.Rotation.Q = " + rot,
            "pl.Base = " + base,
            "_text_ = " + _cmd,
            "Draft.autogroup(_text_)",
            "FreeCAD.ActiveDocument.recompute()",
        ]
        self.commit(translate("draft", "Create Text"), _cmd_list)
        self.finish(cont=None)

    def action(self, arg):
        """Handle the 3D scene events.

        This is installed as an EventCallback in the Inventor view.

        Parameters
        ----------
        arg: dict
            Dictionary with strings that indicates the type of event received
            from the 3D view.
        """
        if arg["Type"] == "SoKeyboardEvent":
            if arg["Key"] == "ESCAPE":
                self.finish()
        elif arg["Type"] == "SoLocation2Event":  # mouse movement detection
            if self.active:
                (self.point, ctrlPoint, info) = gui_tool_utils.getPoint(self, arg)
            gui_tool_utils.redraw3DView()
        elif arg["Type"] == "SoMouseButtonEvent":
            if arg["State"] == "DOWN" and arg["Button"] == "BUTTON1":
                if self.point:
                    self.active = False
                    Gui.Snapper.off()
                    self.node.append(self.point)
                    self.ui.textUi()
                    self.ui.textValue.setFocus()

    def numericInput(self, numx, numy, numz):
        """Validate the entry fields in the user interface.

        This function is called by the toolbar or taskpanel interface
        when valid x, y, and z have been entered in the input fields.
        """
        self.point = App.Vector(numx, numy, numz)
        self.node.append(self.point)
        self.ui.textUi()
        self.ui.textValue.setFocus()


Gui.addCommand("Draft_Text", Text())

## @}



from Base.Metadata import export, constmethod
from Base.Vector import Vector
from Base.Rotation import Rotation as RotationPy
from Geometry import Geometry
from typing import Final, overload, List, Union, Optional, Tuple

@export(
    Twin="GeomCurve",
    TwinPointer="GeomCurve",
    PythonName="Part.Curve",
    FatherInclude="Mod/Part/App/GeometryPy.h",
    Include="Mod/Part/App/Geometry.h",
    Constructor=True,
)
class GeometryCurve(Geometry):
    """
    The abstract class GeometryCurve is the root class of all curve objects.

    Author: Werner Mayer (wmayer@users.sourceforge.net)
    Licence: LGPL
    """

    Continuity: Final[str] = ""
    """Returns the global continuity of the curve."""

    FirstParameter: Final[float] = 0.0
    """Returns the value of the first parameter."""

    LastParameter: Final[float] = 0.0
    """Returns the value of the last parameter."""

    Rotation: Final[RotationPy] = ...
    """Returns a rotation object to describe the orientation for curve that supports it"""

    @constmethod
    def toShape(self) -> object:
        """
        Return the shape for the geometry.
        """
        ...

    @overload
    @constmethod
    def discretize(
        self, Number: int, *, First: Optional[float] = None, Last: Optional[float] = None
    ) -> List[Vector]:
        """
        Discretizes the curve and returns a list of points.
        """
        ...

    @overload
    @constmethod
    def discretize(
        self, QuasiNumber: int, *, First: Optional[float] = None, Last: Optional[float] = None
    ) -> List[Vector]:
        """
        Discretizes the curve and returns a list of quasi equidistant points.
        """
        ...

    @overload
    @constmethod
    def discretize(
        self, Distance: float, *, First: Optional[float] = None, Last: Optional[float] = None
    ) -> List[Vector]:
        """
        Discretizes the curve and returns a list of equidistant points with distance 'd'.
        """
        ...

    @overload
    @constmethod
    def discretize(
        self, Deflection: float, *, First: Optional[float] = None, Last: Optional[float] = None
    ) -> List[Vector]:
        """
        Discretizes the curve and returns a list of points with a maximum deflection 'd' to the curve.
        """
        ...

    @overload
    @constmethod
    def discretize(
        self, QuasiDeflection: float, *, First: Optional[float] = None, Last: Optional[float] = None
    ) -> List[Vector]:
        """
        Discretizes the curve and returns a list of points with a maximum deflection 'd' to the curve (faster).
        """
        ...

    @overload
    @constmethod
    def discretize(
        self,
        Angular: float,
        Curvature: float,
        Minimum: int = 2,
        *,
        First: Optional[float] = None,
        Last: Optional[float] = None
    ) -> List[Vector]:
        """
        Discretizes the curve and returns a list of points with an angular deflection of 'a' and a curvature deflection of 'c'.
        Optionally a minimum number of points can be set.
        """
        ...

    @constmethod
    def discretize(self, **kwargs) -> List[Vector]:
        """
        Discretizes the curve and returns a list of points.

        The function accepts keywords as argument:

        discretize(Number=n) => gives a list of 'n' equidistant points
        discretize(QuasiNumber=n) => gives a list of 'n' quasi equidistant points (is faster than the method above)
        discretize(Distance=d) => gives a list of equidistant points with distance 'd'
        discretize(Deflection=d) => gives a list of points with a maximum deflection 'd' to the curve
        discretize(QuasiDeflection=d) => gives a list of points with a maximum deflection 'd' to the curve (faster)
        discretize(Angular=a,Curvature=c,[Minimum=m]) => gives a list of points with an angular deflection of 'a'
                                            and a curvature deflection of 'c'. Optionally a minimum number of points
                                            can be set which by default is set to 2.

        Optionally you can set the keywords 'First' and 'Last' to define a sub-range of the parameter range
        of the curve.

        If no keyword is given then it depends on whether the argument is an int or float.
        If it's an int then the behaviour is as if using the keyword 'Number', if it's float
        then the behaviour is as if using the keyword 'Distance'.

        Example:

        import Part
        c=Part.Circle()
        c.Radius=5
        p=c.discretize(Number=50,First=3.14)
        s=Part.Compound([Part.Vertex(i) for i in p])
        Part.show(s)


        p=c.discretize(Angular=0.09,Curvature=0.01,Last=3.14,Minimum=100)
        s=Part.Compound([Part.Vertex(i) for i in p])
        Part.show(s)
        """
        ...

    @constmethod
    def getD0(self, parameter: float) -> Vector:
        """
        Returns the point of given parameter
        """
        ...

    @constmethod
    def getD1(self, parameter: float) -> Vector:
        """
        Returns the point and first derivative of given parameter
        """
        ...

    @constmethod
    def getD2(self, parameter: float) -> Vector:
        """
        Returns the point, first and second derivatives
        """
        ...

    @constmethod
    def getD3(self, parameter: float) -> Vector:
        """
        Returns the point, first, second and third derivatives
        """
        ...

    @constmethod
    def getDN(self, n: int, parameter: float) -> Vector:
        """
        Returns the n-th derivative
        """
        ...

    @constmethod
    def length(
        self,
        uMin: Optional[float] = None,
        uMax: Optional[float] = None,
        Tol: Optional[float] = None,
    ) -> float:
        """
        Computes the length of a curve
        length([uMin, uMax, Tol]) -> float
        """
        ...

    @constmethod
    def parameterAtDistance(
        self,
        abscissa: Optional[float] = None,
        startingParameter: Optional[float] = None,
    ) -> float:
        """
        Returns the parameter on the curve of a point at the given distance from a starting parameter.
        parameterAtDistance([abscissa, startingParameter]) -> float
        """
        ...

    @constmethod
    def value(self, u: float) -> Vector:
        """
        Computes the point of parameter u on this curve
        """
        ...

    @constmethod
    def tangent(self, u: float) -> Vector:
        """
        Computes the tangent of parameter u on this curve
        """
        ...

    @constmethod
    def makeRuledSurface(self, otherCurve: "GeometryCurve") -> object:
        """
        Make a ruled surface of this and the given curves
        """
        ...

    @constmethod
    def intersect2d(self, otherCurve: "GeometryCurve") -> List[Vector]:
        """
        Get intersection points with another curve lying on a plane.
        """
        ...

    @constmethod
    def continuityWith(self, otherCurve: "GeometryCurve") -> str:
        """
        Computes the continuity of two curves
        """
        ...

    @constmethod
    def parameter(self, point: Vector) -> float:
        """
        Returns the parameter on the curve of the nearest orthogonal projection of the point.
        """
        ...

    @constmethod
    def normal(self, pos: float) -> Vector:
        """
        Vector = normal(pos) - Get the normal vector at the given parameter [First|Last] if defined
        """
        ...

    @overload
    @constmethod
    def projectPoint(self, Point: Vector, Method: str = "NearestPoint") -> Vector:
        """
        projectPoint(Point=Vector, Method="NearestPoint") -> Vector
        """
        ...

    @overload
    @constmethod
    def projectPoint(self, Point: Vector, Method: str = "LowerDistance") -> float:
        """
        projectPoint(Vector, "LowerDistance") -> float.
        """
        ...

    @overload
    @constmethod
    def projectPoint(self, Point: Vector, Method: str = "LowerDistanceParameter") -> float:
        """
        projectPoint(Vector, "LowerDistanceParameter") -> float.
        """
        ...

    @overload
    @constmethod
    def projectPoint(self, Point: Vector, Method: str = "Distance") -> List[float]:
        """
        projectPoint(Vector, "Distance") -> list of floats.
        """
        ...

    @overload
    @constmethod
    def projectPoint(self, Point: Vector, Method: str = "Parameter") -> List[float]:
        """
        projectPoint(Vector, "Parameter") -> list of floats.
        """
        ...

    @overload
    @constmethod
    def projectPoint(self, Point: Vector, Method: str = "Point") -> List[Vector]:
        """
        projectPoint(Vector, "Point") -> list of points.
        """
        ...

    @constmethod
    def projectPoint(self, **kwargs) -> Union[Vector, float, List[float], List[Vector]]:
        """
        Computes the projection of a point on the curve

        projectPoint(Point=Vector,[Method="NearestPoint"])
        projectPoint(Vector,"NearestPoint") -> Vector
        projectPoint(Vector,"LowerDistance") -> float
        projectPoint(Vector,"LowerDistanceParameter") -> float
        projectPoint(Vector,"Distance") -> list of floats
        projectPoint(Vector,"Parameter") -> list of floats
        projectPoint(Vector,"Point") -> list of points
        """
        ...

    @constmethod
    def curvature(self, pos: float) -> float:
        """
        Float = curvature(pos) - Get the curvature at the given parameter [First|Last] if defined
        """
        ...

    @constmethod
    def centerOfCurvature(self, pos: float) -> Vector:
        """
        Vector = centerOfCurvature(float pos) - Get the center of curvature at the given parameter [First|Last] if defined
        """
        ...

    @constmethod
    def intersect(self, curve_or_surface: object, precision: float) -> object:
        """
        Returns all intersection points and curve segments between the curve and the curve/surface.

        arguments: curve/surface (for the intersection), precision (float)
        """
        ...

    @constmethod
    def intersectCS(self, surface: object) -> object:
        """
        Returns all intersection points and curve segments between the curve and the surface.
        """
        ...

    @constmethod
    def intersectCC(self, otherCurve: "GeometryCurve") -> List[Vector]:
        """
        Returns all intersection points between this curve and the given curve.
        """
        ...

    @constmethod
    def toBSpline(self, points: Tuple[float, float]) -> "BSplineCurve":
        """
        Converts a curve of any type (only part from First to Last) to BSpline curve.
        toBSpline((first: float, last: float)) -> BSplineCurve
        """
        ...

    @constmethod
    def toNurbs(self, points: Tuple[float, float]) -> "NurbsCurve":
        """
        Converts a curve of any type (only part from First to Last) to NURBS curve.
        toNurbs((first: float, last: float)) -> NurbsCurve
        """
        ...

    @constmethod
    def trim(self, points: Tuple[float, float]) -> "TrimmedCurve":
        """
        Returns a trimmed curve defined in the given parameter range.
        trim((first: float, last: float)) -> TrimmedCurve
        """
        ...

    @constmethod
    def approximateBSpline(
        self, Tolerance: float, MaxSegments: int, MaxDegree: int, Order: str = "C2"
    ) -> "BSplineCurve":
        """
        Approximates a curve of any type to a B-Spline curve.
        approximateBSpline(Tolerance, MaxSegments, MaxDegree, [Order='C2']) -> BSplineCurve
        """
        ...

    def reverse(self) -> None:
        """
        Changes the direction of parametrization of the curve.
        """
        ...

    @constmethod
    def reversedParameter(self, U: float) -> float:
        """
        Returns the parameter on the reversed curve for the point of parameter U on this curve.
        """
        ...

    @constmethod
    def isPeriodic(self) -> bool:
        """
        Returns true if this curve is periodic.
        """
        ...

    @constmethod
    def period(self) -> float:
        """
        Returns the period of this curve or raises an exception if it is not periodic.
        """
        ...

    @constmethod
    def isClosed(self) -> bool:
        """
        Returns true if the curve is closed.
        """
        ...



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileNotice: Part of the AddonManager.

# FreeCAD runs InitGui.py automatically during the GUI initialization process by reading the
# file into memory and running `exec` on its contents (so __file__ is not defined directly).

import os
import AddonManager

cwd = os.path.dirname(AddonManager.__file__)
FreeCADGui.addLanguagePath(os.path.join(cwd, "Resources", "translations"))
FreeCADGui.addIconPath(os.path.join(cwd, "Resources", "icons"))
FreeCADGui.addCommand("Std_AddonMgr", AddonManager.CommandAddonManager())
